from fs_uae_launcher.ui.behaviors.amigaenablebehavior import AmigaEnableBehavior
import fsui as fsui
from ...Config import Config
from ...DeviceManager import DeviceManager
from ...I18N import gettext
from ...Signal import Signal
from ..IconButton import IconButton
from ..HelpButton import HelpButton


class InputSelector(fsui.Group):

    def __init__(self, parent, port):
        self.port = port
        self.device_option_key = "joystick_port_{0}".format(port)
        self.mode_option_key = "joystick_port_{0}_mode".format(port)
        self.autofire_mode_option_key = "joystick_port_{0}_autofire".format(
            port)

        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        self.joystick_mode_values = ["nothing", "mouse", "joystick",
                                     "cd32 gamepad"]
        self.joystick_mode_titles = [gettext("Nothing"), gettext("Mouse"),
                                     gettext("Joystick"),
                                     gettext("CD32 Pad")]

        self.mode_choice = fsui.Choice(self, self.joystick_mode_titles)
        AmigaEnableBehavior(self.mode_choice)
        self.layout.add(self.mode_choice)
        self.layout.add_spacer(10)
        # else:
        #     self.mode_choice = None
        if port >= 4:
            self.mode_choice.disable()

        # devices = ["", _("No Host Device"), _("Mouse"),
        #         _("Cursor Keys and Right Ctrl/Alt")]
        # for i, name in enumerate(DeviceManager.get_joystick_names()):
        #     devices.append(name)
        #     if not self.joystick_values_initialized:
        #         self.joystick_values.append(DeviceManager.device_ids[i])
        # self.joystick_values_initialized = True

        self.device_choice = fsui.ComboBox(self, [""], read_only=True)
        AmigaEnableBehavior(self.device_choice)
        self.joystick_values = []
        self.rebuild_device_list()
        self.device_choice.set_index(0)
        # print(self.device_choice.get_index())
        self.layout.add(self.device_choice, expand=True)

        if port < 4:
            self.autofire_button = IconButton(self, "16/lightning_off.png")
            self.autofire_button.activated.connect(self.on_autofire_button)
            self.layout.add(self.autofire_button, margin_left=10)
        else:
            self.autofire_button = None

            self.help_button = HelpButton(
                self, "http://fs-uae.net/custom-joystick-port")
            self.layout.add(self.help_button, margin_left=10)

        self.initialize_from_config()
        self.set_config_handlers()

    def rebuild_device_list(self):
        # self.joystick_values = ["", "none", "mouse", "keyboard"]
        # devices = ["", _("No Host Device"), _("Mouse"),
        #         _("Cursor Keys and Right Ctrl/Alt")]
        self.joystick_values = ["", "none"]
        devices = ["", gettext("No Host Device")]

        for i, name in enumerate(DeviceManager.get_joystick_names()):
            devices.append(gettext(name))
            self.joystick_values.append(DeviceManager.device_ids[i])
        self.device_choice.set_items(devices)

    def initialize_from_config(self):
        self.on_config(
            self.device_option_key, Config.get(self.device_option_key))
        self.on_config(
            self.mode_option_key, Config.get(self.mode_option_key))
        self.on_config(
            self.autofire_mode_option_key,
            Config.get(self.autofire_mode_option_key))

    def set_config_handlers(self):
        if self.mode_choice is not None:
            self.mode_choice.on_change = self.on_mode_change
        self.device_choice.on_change = self.on_device_change
        Config.add_listener(self)
        Signal.add_listener("settings_updated", self)
        Signal.add_listener("device_list_updated", self)

    def on_destroy(self):
        print("on_destroy")
        Config.remove_listener(self)
        Signal.remove_listener("settings_updated", self)
        Signal.remove_listener("device_list_updated", self)

    def on_mode_change(self):
        if self.mode_choice is not None:
            index = self.mode_choice.get_index()
            value = self.joystick_mode_values[index]
            self.set_value_or_default(value)

    def set_value_or_default(self, value):
        if self.port == 0:
            if value == "mouse":
                value = ""
        elif self.port == 1:
            if Config.get("amiga_model").startswith("CD32"):
                default = "cd32 gamepad"
            else:
                default = "joystick"
            if value == default:
                value = ""
        else:
            if value == "nothing":
                value = ""
        if Config.get(self.mode_option_key) != value:
            Config.set(self.mode_option_key, value)

    def on_device_change(self):
        index = self.device_choice.get_index()

        value = self.joystick_values[index]
        for port in range(4):
            if self.port == port:
                continue
            key = "joystick_port_{0}".format(port)
            if Config.get(key) == value:
                Config.set(key, "")
        Config.set(self.device_option_key, value)

    def on_autofire_button(self):
        if Config.get(self.autofire_mode_option_key) == "1":
            Config.set(self.autofire_mode_option_key, "")
        else:
            Config.set(self.autofire_mode_option_key, "1")

    def get_calculated_mode(self, port):
        return DeviceManager.get_calculated_port_mode(Config, port)

    def on_config(self, key, value):
        if key == "amiga_model":
            value = Config.get("joystick_port_{0}_mode".format(self.port))
            self.set_value_or_default(value)

        if key == self.mode_option_key or key == "amiga_model":
            value = self.get_calculated_mode(self.port)
            for i, config in enumerate(self.joystick_mode_values):
                if config == value:
                    if self.mode_choice is not None:
                        self.mode_choice.set_index(i)
                        if self.port >= 4:
                            self.device_choice.enable(i != 0)
                    break
            else:
                print("FIXME: could not set mode")
        elif key == self.device_option_key or key == "amiga_model":
            # print(self.joystick_values)
            value_lower = value.lower()
            for i, name in enumerate(self.joystick_values):
                if value_lower == name.lower():
                    self.device_choice.set_index(i)
                    break
        elif key == self.autofire_mode_option_key:
            if self.autofire_button is not None:
                if value == "1":
                    self.autofire_button.set_tooltip(
                        gettext("Auto-Fire is On"))
                    self.autofire_button.set_icon_name("16/lightning_red.png")
                else:
                    self.autofire_button.set_tooltip(
                        gettext("Auto-Fire is Off"))
                    self.autofire_button.set_icon_name("16/lightning_off.png")

        # this is intended to catch all config changes for all ports (both
        # mode and device) to update the defaults
        if key.startswith("joystick_port_") or key == "amiga_model":
            self.update_default_device()

    def on_device_list_updated_signal(self):
        # print(self.device_choice.get_index())
        had_default = (self.device_choice.get_index() == 0)
        self.rebuild_device_list()
        self.update_default_device(had_default=had_default)

    def on_settings_updated_signal(self):
        self.update_default_device()

    def update_default_device(self, had_default=None):
        config = {}
        for port in range(4):
            key = "joystick_port_{0}".format(port)
            if self.port == port:
                config[key] = ""
            else:
                config[key] = Config.get(key)
            key = "joystick_port_{0}_mode".format(port)
            config[key] = self.get_calculated_mode(port)
        device = DeviceManager.get_device_for_port(config, self.port)
        default_description = gettext("Default ({0})").format(
            gettext(device.name))
        # print("default_description = ", default_description)

        if had_default is None:
            had_default = (self.device_choice.get_index() == 0)
        # print("had default", had_default, self.device_choice.get_index())
        self.device_choice.set_item_text(0, default_description)
        # print("had_default", had_default)
        if had_default:
            # print("set text for", self.port, default_description)
            # self.device_choice.set_index(1)
            self.device_choice.set_text(default_description)
            self.device_choice.set_index(0)
        # print(self.device_choice.get_index())



import fsui
from fsgs.context import fsgs
from ...Config import Config


class CustomOptionsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        # self.layout.add_spacer(20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True, expand=True)

        # hor_layout.add_spacer(20)
        self.text_area = fsui.TextArea(self, font_family="monospace")
        self.text_area.set_min_width(760)
        self.text_area.set_min_height(460)
        self.text_area.set_text(self.get_initial_text())
        hor_layout.add(self.text_area, fill=True, expand=True)
        # hor_layout.add_spacer(20)

        # self.layout.add_spacer(20)

        self.get_window().add_close_listener(self.on_close_window)

    def on_close_window(self):
        self.update_config()

    def on_close_button(self):
        self.end_modal(0)

    def update_config(self):
        text = self.text_area.get_text()
        for key in list(fsgs.config.values.keys()):
            if key not in Config.default_config:
                del fsgs.config.values[key]

        for line in text.split("\n"):
            line = line.strip()
            if line.startswith("# You can write key = value pairs here"):
                continue
            parts = line.split("=", 1)
            if len(parts) == 2:
                key = parts[0].strip()
                # if key in Config.no_custom_config:
                #     continue
                value = parts[1].strip()
                Config.set(key, value)

    def get_initial_text(self):
        text = DEFAULT_TEXT
        keys = fsgs.config.values.keys()
        for key in sorted(keys):
            if key in Config.no_custom_config:
                continue
            value = fsgs.config.values[key]
            if not value:
                continue
            text += "{0} = {1}\n".format(key, value)
        return text

DEFAULT_TEXT = """\
# You can write key = value pairs here to set FS-UAE options
# not currently supported by the user interface. This is only a
# temporary feature until the GUI supports all options directly.
#
# The options specified here will apply to this configuration only.

"""



#!/usr/bin/env python3

import sys

if sys.version_info[0] < 3 or sys.version_info[1] < 2:
    print("You need at least Python 3.2 to run FS-UAE Launcher")
    sys.exit(1)

# Workaround to make import typing work without having it on the default
# python path (would confuse mypy).

import fstd.typing
sys.modules["typing"] = fstd.typing

if "--server" in sys.argv:
    sys.argv.remove("--server")
    app = "fs-uae-netplay-server"
elif "--fs-uae-arcade" in sys.argv:
    sys.argv.remove("--fs-uae-arcade")
    app = "fs-uae-arcade"
elif "--fs-game-center" in sys.argv:
    sys.argv.remove("--fs-game-center")
    app = "fs-game-center"
elif sys.argv[0].endswith("fs-game-center"):
    app = "fs-game-center"
else:
    app = "fs-uae-launcher"


def main():
    version = "2.6.2"

    # if "--fs-uae-workspace=real" in sys.argv:
    #     try:
    #         # noinspection PyUnresolvedReferences
    #         from gi.repository import Gdk
    #     except ImportError:
    #         Gdk = None

    for arg in sys.argv:
        # if not isinstance(arg, unicode):
        if isinstance(arg, bytes):
            arg = arg.decode(sys.getfilesystemencoding())
        if arg.startswith("--") and "=" in arg:
            key, value = arg[2:].split("=", 1)
            key = key.replace("-", "_")
            if key == "fake_version":
                version = value

    from fsgs.FSGSDirectories import FSGSDirectories
    FSGSDirectories.initialize()

    import socket
    socket.setdefaulttimeout(30.0)

    from fsbc.init import initialize_application
    initialize_application(app, version=version)

    if app == "fs-uae-netplay-server":
        from fs_uae_launcher.server.game import main
        main()

    elif app == "fs-game-center":
        print("FS Game Center")

        from fsui import Application
        application = Application("fs-game-center")

        from fsgs.http.server import http_server_main
        import threading

        def http_server_thread():
            return http_server_main()
        
        threading.Thread(target=http_server_thread).start()

        # from fsgs.ui.mainwindow import FSGSMainWindow
        # window = FSGSMainWindow()
        from fsgs.ui.qwindow import GameCenterView
        window = GameCenterView()
        from fsui.qt import Qt
        # window.setFlags(Qt.FramelessWindowHint)
        window.setFlags(Qt.Window | 
                        Qt.FramelessWindowHint | 
                        Qt.WindowMinimizeButtonHint | 
                        Qt.WindowSystemMenuHint)
        # window.show()
        window.showMaximized()
        # window.showFullScreen()
        # window.setWindowState(Qt.WindowFullScreen)
        # window.setVisible(True)

        # from fs_uae_launcher.FSUAELauncher import FSUAELauncher
        # application = FSUAELauncher()

        # if application.start():
        application.run()
        # application.save_settings()

        from fsbc.signal import Signal
        Signal("quit").notify()

    elif app == "fs-uae-arcade":
        try:
            import game_center.main
            game_center.main.main()
            print("main returned")

        finally:
            from fsbc.Application import Application
            application = Application.instance()
            if application:
                print("calling Application stop")
                Application.get().stop()

            from fsbc.signal import Signal
            print("sending quit signal")
            Signal("quit").notify()

    else:
        print("FS-UAE Launcher {0}".format(version))

        # if "--workspace" in sys.argv:
        #     if "--qt" not in sys.argv:
        #         sys.argv.insert(1, "--qt")

        from fs_uae_launcher.ConfigChecker import ConfigChecker
        ConfigChecker()

        from fs_uae_launcher.FSUAELauncher import FSUAELauncher
        application = FSUAELauncher()

        if "--fs-uae-workspace=real" in sys.argv:
            from fs_uae_workspace.realdesktop.taskbarwindow import TaskBarWindow
            from fs_uae_workspace.realdesktop.titlebarwindow import TitleBarWindow
            # from fs_uae_workspace.realdesktop.realdesktopwindow import RealDesktopWindow
            from fs_uae_workspace.realdesktop.realwindowhandler import RealWindowHandler
            # from fs_uae_workspace.shellui.shellwidget import ShellWidget

            from fs_uae_workspace.shellui.taskbarwidget import TaskBarWidget
            from fs_uae_workspace.shellui.titlebarwidget import TitleBarWidget
            # from fs_uae_workspace.shellui.desktopwidget import DesktopWidget

            window_handler = RealWindowHandler()

            title_bar_window = TitleBarWindow()
            title_bar_widget = TitleBarWidget(title_bar_window, window_handler)
            title_bar_window.setCentralWidget(title_bar_widget)
            title_bar_window.show()

            task_bar_window = TaskBarWindow()
            task_bar_widget = TaskBarWidget(task_bar_window, window_handler)
            task_bar_window.setCentralWidget(task_bar_widget)
            task_bar_window.show()

            # desktop_window = RealDesktopWindow()
            # shell_widget = ShellWidget(desktop_window, window_handler)
            # desktop_window.set_shell_widget(shell_widget)

            # desktop_window.show()
            application.run()

            sys.exit(0)

        elif "--workspace" in sys.argv:
            if ":" in sys.argv[-1]:
                from fs_uae_workspace.shell import shell_open
                shell_open(sys.argv[-1])

        if application.start():
            application.run()
        application.save_settings()

        from fs_uae_launcher.netplay.IRC import IRC
        IRC.stop()

        from fsbc.signal import Signal
        Signal("quit").notify()



import fsui as fsui
from ..Scanner import Scanner
from ..I18N import gettext
from .settings.ScanPathsGroup import ScanPathsGroup


TIMER_INTERVAL = 100


class ScanDialog(fsui.Dialog):

    @classmethod
    def refresh_game_database(cls, window):
        return cls(window, minimal=True, interactive=False,
                   scan_for_files=False)

    def __init__(self, parent, minimal=False, interactive=True,
                 scan_for_files=True):
        super().__init__(parent, gettext("Scan"))
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        buttons.create_close_button()

        self.layout.add_spacer(640, 0)

        self.interactive = interactive
        self.scan_for_files = scan_for_files
        self.update_game_database = False
        # self.update_game_database = True

        if not minimal:
            # layout.add_spacer(20)

            from .ScanKickstartGroup import ScanKickstartGroup
            self.scan_kickstart_group = ScanKickstartGroup(self)
            layout.add(self.scan_kickstart_group, fill=True)

            layout.add_spacer(20)

            label = fsui.HeadingLabel(
                self, gettext("Scan for Kickstarts, Files and Configurations"))
            layout.add(label, margin_bottom=10)

            self.scan_paths_group = ScanPathsGroup(self)
            layout.add(self.scan_paths_group, fill=True, margin=0)

            layout.add_spacer(20)

        from .ScanProgressGroup import ScanProgressGroup
        self.scan_progress_group = ScanProgressGroup(self)
        layout.add(self.scan_progress_group, fill=True)

        # layout.add_spacer(20)
        # layout.add_spacer(20)

        # hor_layout = fsui.HorizontalLayout()
        # layout.add(hor_layout, fill=True)

        # hor_layout.add_spacer(10, expand=True)
        if interactive:
            self.scan_button = buttons.add_button(
                fsui.Button(buttons, gettext("Scan")))
            # self.scan_button = fsui.Button(self, _("Scan"))
            # hor_layout.add(self.scan_button)
            # hor_layout.add_spacer(10)
            self.scan_button.activated.connect(self.on_scan_button)
        else:
            self.scan_button = None

        # self.stop_button = fsui.Button(self, _("Abort"))
        # hor_layout.add(self.stop_button)
        self.stop_button = buttons.add_button(
            fsui.Button(buttons, gettext("Abort")))
        self.stop_button.activated.connect(self.on_stop_button)

        # hor_layout.add_spacer(10)
        # self.close_button = fsui.Button(self, _("Close"))
        # self.close_button.activated.connect(self.on_close_button
        # hor_layout.add(self.close_button)
        # hor_layout.add_spacer(10)

        # layout.add_spacer(10)

        # self.set_size(layout.get_min_size())
        # self.center_on_parent()

        self.old_title = ""
        self.old_status = ""
        self.has_started_scan = False

        # self.on_timer()
        # self.set_timer(TIMER_INTERVAL)
        self.timer = fsui.IntervalTimer(TIMER_INTERVAL)
        self.timer.activated.connect(self.on_timer)

        if not self.interactive:
            self.start_scan()

    def __del__(self):
        print("ScanDialog.__del__")

    def on_close(self):
        Scanner.stop_flag = True

    def on_destroy(self):
        Scanner.stop_flag = True

    def set_scan_title(self, text):
        if not text:
            return
        if text == self.old_title:
            return
        self.old_title = text
        self.scan_progress_group.title_label.set_text(text)

    def set_scan_status(self, text):
        if not text:
            return
        if text == self.old_status:
            return
        self.old_status = text
        self.scan_progress_group.status_label.set_text(text)

    def on_timer(self):
        if not Scanner.running:
            if self.has_started_scan:
                if Scanner.error:
                    self.set_scan_title(gettext("Scan error"))
                    self.set_scan_status(Scanner.error)
                else:
                    if not self.interactive:
                        self.end_modal(True)
                        return
                    self.set_scan_title(gettext("Scan complete"))
                    self.set_scan_status(
                        gettext("Click 'Scan' button if you want to re-scan"))
            else:
                self.set_scan_title(gettext("No scan in progress"))
                self.set_scan_status(
                    gettext("Click 'Scan' button to start scan"))
            if self.scan_button is not None:
                self.scan_button.enable()
            self.stop_button.disable()
            # self.close_button.enable()
            return

        status = Scanner.status
        self.set_scan_title(status[0])
        self.set_scan_status(status[1])

    def on_scan_button(self):
        self.start_scan()

    def start_scan(self):
        if self.scan_button is not None:
            self.scan_button.disable()
        self.has_started_scan = True
        self.set_scan_title(gettext("Starting scan"))
        self.set_scan_status(gettext("Please wait..."))
        paths = ScanPathsGroup.get_search_path()

        # self.close_button.disable()
        self.stop_button.enable()

        Scanner.start(paths, scan_for_files=self.scan_for_files,
                      update_game_database=self.update_game_database,
                      purge_other_dirs=True)

    # def on_close_button(self):
    #     self.end_modal(False)

    def on_stop_button(self):
        Scanner.stop_flag = True
        # self.close_button.enable()



from fsbc.system import macosx


class InputMapper(object):

    def __init__(self, input, mapping):
        self.input = input
        self.device = input.device
        self.mapping = mapping

    def items(self):
        # input = self.input
        # if not input.device_config:
        #     return
        if not self.device:
            return
        config = self.device.configure(self.input.mapping_name)
        # print(config)
        print("mapping:", self.mapping)
        for native_button, game_button in config.items():
            game_button = game_button.upper()
            print("******", native_button, game_button)
            try:
                input_key = self.mapping[game_button]
            except KeyError:
                pass
            else:
                print("input_key", input_key)
                input_value = self.calc_input(native_button)
                # print("---->", input_key, input_value)
                # f.write('{input_key} = "{input_value}"\n'.format(
                #         input_key=input_key,
                #         input_value=input_value))
                if input_value is not None:
                    yield input_key, input_value

    def axis(self, axis, positive):
        raise NotImplementedError()

    def hat(self, hat, direction):
        raise NotImplementedError()

    def button(self, button):
        raise NotImplementedError()

    def key(self, key):
        raise NotImplementedError()

    def mouse(self, button, axis, positive):
        raise NotImplementedError()

    def calc_input(self, value):
        parts = value.lower().split('_')
        if parts[0] == 'axis':
            axis = int(parts[1])
            positive = parts[2] == 'pos'
            return self.axis(axis, positive)
        elif parts[0] == 'hat':
            hat = int(parts[1])
            direction = parts[2]
            return self.hat(hat, direction)
        elif parts[0] == 'button':
            button = int(parts[1])
            return self.button(button)
        elif parts[0] == 'key':
            key_name = value.split('_', 1)[1].lower()
            if key_name == "rctrl" and macosx:
                print("using ralt instead of rctrl on Mac")
                key_name = "ralt"
            key = Keyboard.key(key_name)
            return self.key(key)
        elif parts[0] == 'mouse':
            button = 0
            axis = 0
            positive = False
            if parts[1] == 'left':
                button = 1
            elif parts[1] == 'right':
                button = 2
            elif parts[1] == 'middle':
                button = 3
            else:
                axis = int(parts[1])
                positive = parts[2] == 'pos'
            return self.mouse(button, axis, positive)
        return None
        # raise Exception("Invalid input value: {0}".format(value))


class Key(object):

    def __init__(self, name):
        self.name = name

    @property
    def sdl_name(self):
        return self.name

    @property
    def sdl_code(self):
        return key_table[self.name][0]

    @property
    def dinput_code(self):
        return key_table[self.name][2]

    @property
    def dinput_name(self):
        return key_table[self.name][1]


class Keyboard(object):

    @staticmethod
    def key(name):
        if isinstance(name, int):
            name = sdl_key_code_to_name[name]
        name = name.upper()
        if name.startswith('SDLK_'):
            pass
        else:
            name = 'SDLK_' + name
        return Key(name)


key_table = {
    'SDLK_0': (48, 'DIK_0', 11),
    'SDLK_1': (49, 'DIK_1', 2),
    'SDLK_2': (50, 'DIK_2', 3),
    'SDLK_3': (51, 'DIK_3', 4),
    'SDLK_4': (52, 'DIK_4', 5),
    'SDLK_5': (53, 'DIK_5', 6),
    'SDLK_6': (54, 'DIK_6', 7),
    'SDLK_7': (55, 'DIK_7', 8),
    'SDLK_8': (56, 'DIK_8', 9),
    'SDLK_9': (57, 'DIK_9', 10),
    'SDLK_A': (97, 'DIK_A', 30),
    'SDLK_AMPERSAND': (38, '', 0),
    'SDLK_ASTERISK': (42, '', 0),
    'SDLK_AT': (64, 'DIK_AT', 145),
    'SDLK_B': (98, 'DIK_B', 48),
    'SDLK_BACKQUOTE': (96, 'DIK_GRAVE', 0),
    'SDLK_BACKSLASH': (92, 'DIK_BACKSLASH', 43),
    'SDLK_BACKSPACE': (8, 'DIK_BACK', 14),
    'SDLK_BREAK': (318, '', 0),
    'SDLK_C': (99, 'DIK_C', 46),
    'SDLK_CAPSLOCK': (301, 'DIK_CAPITAL', 58),
    'SDLK_CARET': (94, '', 0),
    'SDLK_CLEAR': (12, '', 0),
    'SDLK_COLON': (58, 'DIK_COLON', 146),
    'SDLK_COMMA': (44, 'DIK_COMMA', 51),
    'SDLK_COMPOSE': (314, '', 0),
    'SDLK_D': (100, 'DIK_D', 32),
    'SDLK_DELETE': (127, 'DIK_DELETE', 211),
    'SDLK_DOLLAR': (36, '', 0),
    'SDLK_DOWN': (274, 'DIK_DOWN', 208),
    'SDLK_E': (101, 'DIK_E', 18),
    'SDLK_END': (279, 'DIK_END', 207),
    'SDLK_EQUALS': (61, 'DIK_EQUALS', 13),
    'SDLK_ESCAPE': (27, 'DIK_ESCAPE', 1),
    'SDLK_EURO': (321, '', 0),
    'SDLK_EXCLAIM': (33, '', 0),
    'SDLK_F': (102, 'DIK_F', 33),
    'SDLK_F1': (282, 'DIK_F1', 59),
    'SDLK_F10': (291, 'DIK_F10', 68),
    'SDLK_F11': (292, 'DIK_F11', 87),
    'SDLK_F12': (293, 'DIK_F12', 88),
    'SDLK_F13': (294, 'DIK_F13', 100),
    'SDLK_F14': (295, 'DIK_F14', 101),
    'SDLK_F15': (296, 'DIK_F15', 102),
    'SDLK_F2': (283, 'DIK_F2', 60),
    'SDLK_F3': (284, 'DIK_F3', 61),
    'SDLK_F4': (285, 'DIK_F4', 62),
    'SDLK_F5': (286, 'DIK_F5', 63),
    'SDLK_F6': (287, 'DIK_F6', 64),
    'SDLK_F7': (288, 'DIK_F7', 65),
    'SDLK_F8': (289, 'DIK_F8', 66),
    'SDLK_F9': (290, 'DIK_F9', 67),
    'SDLK_G': (103, 'DIK_G', 34),
    'SDLK_GREATER': (62, '', 0),
    'SDLK_H': (104, 'DIK_H', 35),
    'SDLK_HASH': (35, '', 0),
    'SDLK_HELP': (315, '', 0),
    'SDLK_HOME': (278, 'DIK_HOME', 199),
    'SDLK_I': (105, 'DIK_I', 23),
    'SDLK_INSERT': (277, 'DIK_INSERT', 210),
    'SDLK_J': (106, 'DIK_J', 36),
    'SDLK_K': (107, 'DIK_K', 37),
    'SDLK_KP0': (256, 'DIK_NUMPAD0', 82),
    'SDLK_KP1': (257, 'DIK_NUMPAD1', 79),
    'SDLK_KP2': (258, 'DIK_NUMPAD2', 80),
    'SDLK_KP3': (259, 'DIK_NUMPAD3', 81),
    'SDLK_KP4': (260, 'DIK_NUMPAD4', 75),
    'SDLK_KP5': (261, 'DIK_NUMPAD5', 76),
    'SDLK_KP6': (262, 'DIK_NUMPAD6', 77),
    'SDLK_KP7': (263, 'DIK_NUMPAD7', 71),
    'SDLK_KP8': (264, 'DIK_NUMPAD8', 72),
    'SDLK_KP9': (265, 'DIK_NUMPAD9', 73),
    'SDLK_KP_DIVIDE': (267, 'DIK_DIVIDE', 181),
    'SDLK_KP_ENTER': (271, 'DIK_NUMPADENTER', 156),
    'SDLK_KP_EQUALS': (272, '', 141),
    'SDLK_KP_MINUS': (269, 'DIK_SUBTRACT', 74),
    'SDLK_KP_MULTIPLY': (268, 'DIK_MULTIPLY', 55),
    'SDLK_KP_PERIOD': (266, 'DIK_DECIMAL', 83),
    'SDLK_KP_PLUS': (270, 'DIK_ADD', 78),
    'SDLK_L': (108, 'DIK_L', 38),
    'SDLK_LALT': (308, 'DIK_LMENU', 56),
    'SDLK_LCTRL': (306, 'DIK_LCONTROL', 29),
    'SDLK_LEFT': (276, 'DIK_LEFT', 203),
    'SDLK_LEFTBRACKET': (91, 'DIK_LBRACKET', 26),
    'SDLK_LEFTPAREN': (40, '', 0),
    'SDLK_LESS': (60, '', 0),
    'SDLK_LMETA': (310, '', 0),
    'SDLK_LSHIFT': (304, 'DIK_LSHIFT', 42),
    'SDLK_LSUPER': (311, 'DIK_LWIN', 219),
    'SDLK_M': (109, 'DIK_M', 50),
    'SDLK_MENU': (319, '', 0),
    'SDLK_MINUS': (45, 'DIK_MINUS', 12),
    'SDLK_MODE': (313, '', 0),
    'SDLK_N': (110, 'DIK_N', 49),
    'SDLK_NUMLOCK': (300, 'DIK_NUMLOCK', 69),
    'SDLK_O': (111, 'DIK_O', 24),
    'SDLK_P': (112, 'DIK_P', 25),
    'SDLK_PAGEDOWN': (281, 'DIK_NEXT', 209),
    'SDLK_PAGEUP': (280, 'DIK_PRIOR', 201),
    'SDLK_PAUSE': (19, '', 0),
    'SDLK_PERIOD': (46, 'DIK_PERIOD', 52),
    'SDLK_PLUS': (43, '', 0),
    'SDLK_POWER': (320, '', 0),
    'SDLK_PRINT': (316, '', 0),
    'SDLK_Q': (113, 'DIK_Q', 16),
    'SDLK_QUESTION': (63, '', 0),
    'SDLK_QUOTE': (39, 'DIK_APOSTROPHE', 40),
    'SDLK_QUOTEDBL': (34, '', 0),
    'SDLK_R': (114, 'DIK_R', 19),
    'SDLK_RALT': (307, 'DIK_RMENU', 184),
    'SDLK_RCTRL': (305, 'DIK_RCONTROL', 157),
    'SDLK_RETURN': (13, 'DIK_RETURN', 28),
    'SDLK_RIGHT': (275, 'DIK_RIGHT', 205),
    'SDLK_RIGHTBRACKET': (93, 'DIK_RBRACKET', 27),
    'SDLK_RIGHTPAREN': (41, '', 0),
    'SDLK_RMETA': (309, '', 0),
    'SDLK_RSHIFT': (303, 'DIK_RSHIFT', 54),
    'SDLK_RSUPER': (312, 'DIK_RWIN', 220),
    'SDLK_S': (115, 'DIK_S', 31),
    'SDLK_SCROLLOCK': (302, 'DIK_SCROLL', 70),
    'SDLK_SEMICOLON': (59, 'DIK_SEMICOLON', 39),
    'SDLK_SLASH': (47, 'DIK_SLASH', 53),
    'SDLK_SPACE': (32, 'DIK_SPACE', 57),
    'SDLK_SYSREQ': (317, 'DIK_SYSRQ', 183),
    'SDLK_T': (116, 'DIK_T', 20),
    'SDLK_TAB': (9, 'DIK_TAB', 15),
    'SDLK_U': (117, 'DIK_U', 22),
    'SDLK_UNDERSCORE': (95, '', 147),
    'SDLK_UNDO': (322, '', 0),
    'SDLK_UP': (273, 'DIK_UP', 200),
    'SDLK_V': (118, 'DIK_V', 47),
    'SDLK_W': (119, 'DIK_W', 17),
    'SDLK_WORLD_0': (160, '', ),
    'SDLK_WORLD_1': (161, '', ),
    'SDLK_WORLD_10': (170, '', ),
    'SDLK_WORLD_11': (171, '', ),
    'SDLK_WORLD_12': (172, '', ),
    'SDLK_WORLD_13': (173, '', ),
    'SDLK_WORLD_14': (174, '', ),
    'SDLK_WORLD_15': (175, '', ),
    'SDLK_WORLD_16': (176, '', ),
    'SDLK_WORLD_17': (177, '', ),
    'SDLK_WORLD_18': (178, '', ),
    'SDLK_WORLD_19': (179, '', ),
    'SDLK_WORLD_2': (162, '', ),
    'SDLK_WORLD_20': (180, '', ),
    'SDLK_WORLD_21': (181, '', ),
    'SDLK_WORLD_22': (182, '', ),
    'SDLK_WORLD_23': (183, '', ),
    'SDLK_WORLD_24': (184, '', ),
    'SDLK_WORLD_25': (185, '', ),
    'SDLK_WORLD_26': (186, '', ),
    'SDLK_WORLD_27': (187, '', ),
    'SDLK_WORLD_28': (188, '', ),
    'SDLK_WORLD_29': (189, '', ),
    'SDLK_WORLD_3': (163, '', ),
    'SDLK_WORLD_30': (190, '', ),
    'SDLK_WORLD_31': (191, '', ),
    'SDLK_WORLD_32': (192, '', ),
    'SDLK_WORLD_33': (193, '', ),
    'SDLK_WORLD_34': (194, '', ),
    'SDLK_WORLD_35': (195, '', ),
    'SDLK_WORLD_36': (196, '', ),
    'SDLK_WORLD_37': (197, '', ),
    'SDLK_WORLD_38': (198, '', ),
    'SDLK_WORLD_39': (199, '', ),
    'SDLK_WORLD_4': (164, '', ),
    'SDLK_WORLD_40': (200, '', ),
    'SDLK_WORLD_41': (201, '', ),
    'SDLK_WORLD_42': (202, '', ),
    'SDLK_WORLD_43': (203, '', ),
    'SDLK_WORLD_44': (204, '', ),
    'SDLK_WORLD_45': (205, '', ),
    'SDLK_WORLD_46': (206, '', ),
    'SDLK_WORLD_47': (207, '', ),
    'SDLK_WORLD_48': (208, '', ),
    'SDLK_WORLD_49': (209, '', ),
    'SDLK_WORLD_5': (165, '', ),
    'SDLK_WORLD_50': (210, '', ),
    'SDLK_WORLD_51': (211, '', ),
    'SDLK_WORLD_52': (212, '', ),
    'SDLK_WORLD_53': (213, '', ),
    'SDLK_WORLD_54': (214, '', ),
    'SDLK_WORLD_55': (215, '', ),
    'SDLK_WORLD_56': (216, '', ),
    'SDLK_WORLD_57': (217, '', ),
    'SDLK_WORLD_58': (218, '', ),
    'SDLK_WORLD_59': (219, '', ),
    'SDLK_WORLD_6': (166, '', ),
    'SDLK_WORLD_60': (220, '', ),
    'SDLK_WORLD_61': (221, '', ),
    'SDLK_WORLD_62': (222, '', ),
    'SDLK_WORLD_63': (223, '', ),
    'SDLK_WORLD_64': (224, '', ),
    'SDLK_WORLD_65': (225, '', ),
    'SDLK_WORLD_66': (226, '', ),
    'SDLK_WORLD_67': (227, '', ),
    'SDLK_WORLD_68': (228, '', ),
    'SDLK_WORLD_69': (229, '', ),
    'SDLK_WORLD_7': (167, '', ),
    'SDLK_WORLD_70': (230, '', ),
    'SDLK_WORLD_71': (231, '', ),
    'SDLK_WORLD_72': (232, '', ),
    'SDLK_WORLD_73': (233, '', ),
    'SDLK_WORLD_74': (234, '', ),
    'SDLK_WORLD_75': (235, '', ),
    'SDLK_WORLD_76': (236, '', ),
    'SDLK_WORLD_77': (237, '', ),
    'SDLK_WORLD_78': (238, '', ),
    'SDLK_WORLD_79': (239, '', ),
    'SDLK_WORLD_8': (168, '', ),
    'SDLK_WORLD_80': (240, '', ),
    'SDLK_WORLD_81': (241, '', ),
    'SDLK_WORLD_82': (242, '', ),
    'SDLK_WORLD_83': (243, '', ),
    'SDLK_WORLD_84': (244, '', ),
    'SDLK_WORLD_85': (245, '', ),
    'SDLK_WORLD_86': (246, '', ),
    'SDLK_WORLD_87': (247, '', ),
    'SDLK_WORLD_88': (248, '', ),
    'SDLK_WORLD_89': (249, '', ),
    'SDLK_WORLD_9': (169, '', ),
    'SDLK_WORLD_90': (250, '', ),
    'SDLK_WORLD_91': (251, '', ),
    'SDLK_WORLD_92': (252, '', ),
    'SDLK_WORLD_93': (253, '', ),
    'SDLK_WORLD_94': (254, '', ),
    'SDLK_WORLD_95': (255, '', ),
    'SDLK_X': (120, 'DIK_X', 45),
    'SDLK_Y': (121, 'DIK_Y', 21),
    'SDLK_Z': (122, 'DIK_Z', 44),
}

sdl_key_code_to_name = {}
for __key, __value in key_table.items():
    sdl_key_code_to_name[__value[0]] = __key



from fsgs.mednafen.mednafen import MednafenRunner


class TurboGrafx16Runner(MednafenRunner):

    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "turbografx16",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "1": "pce.input.port1.gamepad.i",
                "2": "pce.input.port1.gamepad.ii",
                "UP": "pce.input.port1.gamepad.up",
                "DOWN": "pce.input.port1.gamepad.down",
                "LEFT": "pce.input.port1.gamepad.left",
                "RIGHT": "pce.input.port1.gamepad.right",
                "SELECT": "pce.input.port1.gamepad.select",
                "RUN": "pce.input.port1.gamepad.run",
            }
        elif port == 1:
            return {
                "1": "pce.input.port2.gamepad.i",
                "2": "pce.input.port2.gamepad.ii",
                "UP": "pce.input.port2.gamepad.up",
                "DOWN": "pce.input.port2.gamepad.down",
                "LEFT": "pce.input.port2.gamepad.left",
                "RIGHT": "pce.input.port2.gamepad.right",
                "SELECT": "pce.input.port2.gamepad.select",
                "RUN": "pce.input.port2.gamepad.run",
            }

    def mednafen_system_prefix(self):
        return "pce"

    # def medmafen_video_size(self):
    #     # FIXME
    #     return 256, 240



import weakref
from fsui.qt import Qt, QSize, QAbstractListModel
from fsui.qt import QListView
from fsui.qt import QFrame
import fsui

from fsui.qt.Widget import Widget


class Model(QAbstractListModel):

    def __init__(self, parent):
        QAbstractListModel.__init__(self, parent)
        self.parent = weakref.ref(parent)
        self.count = 0

    # def set_item_count(self, count):
    #     self.count = count

    def rowCount(self, parent):
        # print("returning count", self.count, "for parent", parent)
        # return self.count
        return self.parent().get_item_count()

    def data(self, index, role):
        row = index.row()
        # print("data for", index, "role", role)
        # if role == Qt.SizeHintRole:
        #     return QSize(26, 26)
        if role == Qt.DecorationRole:
            icon = self.parent().get_item_icon(row)
            if icon:
                return icon.qpixmap
            # return icon
        elif role == Qt.DisplayRole:
            return self.parent().get_item_text(row)
        # return QVariant()


class IconView(QListView, Widget):

    def __init__(self, parent):
        if hasattr(parent, "get_container"):
            QListView.__init__(self, parent.get_container())
        else:
            QListView.__init__(self, parent)

        self.setViewMode(QListView.IconMode)
        self.setMovement(QListView.Free)
        self.setFlow(QListView.LeftToRight)
        self.setIconSize(QSize(48, 48))
        self.setWrapping(True)
        self.setWordWrap(True)
        # self.setGridSize(QSize(100, 100))
        self.setFrameStyle(QFrame.NoFrame)
        self.setStyleSheet("QListView {background-color: #aaaaaa; }")

        self.setSpacing(30)
        # self.setAutoFillBackground(True)
        # p = self.palette()
        # p.setColor(self.backgroundRole(), QColor(0xaa, 0xaa, 0xff))
        # self.setPalette(p)

        # self.setSelectionModel()
        # self.model = QStandardItemModel(self)
        self.model = Model(self)
        self.setModel(self.model)
        # self.itemSelectionChanged.connect(self.__selection_changed)
        selection_model = self.selectionModel()

        selection_model.selectionChanged.connect(self.__selection_changed)
        self.activated.connect(self.__activated_signal)
        self.resize(600, 400)

    def __activated_signal(self, index):
        index = index.row()
        self.on_activate_item(index)

    def __selection_changed(self):
        indices = self.selectionModel().selectedIndexes()
        if len(indices) == 0:
            print("FIXME: deselect")
        else:
            model_index = indices[0]
            self.on_select_item(model_index.row())

    def get_item_icon(self, index):
        # return QPixmap(48, 48)
        # return QPixmap("/home/frode/git/fs-uae/fs-uae-launcher/icon/fs-uae"
        #                "-launcher/48.png")
        return fsui.Image("")

    def get_item_text(self, index):
        return ""

    def set_default_icon(self, image):
        pass

    def set_index(self, index):
        if index is None:
            index = -1
        # print(self.rootIndex)
        # idx = QModelIndex.createIndex(index)
        idx = self.model.index(index, 0)
        self.setCurrentIndex(idx)

    def select_item(self, index):
        self.set_index(index)

    def on_select_item(self, index):
        pass

    def on_activate_item(self, index):
        pass

    def update(self):
        count = self.get_item_count()
        self.dataChanged(
            self.model.createIndex(0, 0),
            self.model.createIndex(count, 0))



import os
import shutil
from .signal import Signal
from configparser import ConfigParser, NoSectionError
# noinspection PyUnresolvedReferences
from typing import Dict, Tuple


class Settings(object):

    def __init__(self, app=None, path: str=None) -> None:
        self.app = app
        self.path = path
        self.values = {}  # type: Dict[str, str]
        self.load()

    def get(self, key: str) -> str:
        return self.values.get(key, "")

    def set(self, key: str, value: str) -> None:
        if self[key] == value:
            print("settings set {0} to {1} (no change)".format(key, value))
            return
        _log_key_value(key, value)
        self.values[key] = value
        Signal("setting").notify(key, value)

    def __getitem__(self, key: str):
        return self.get(key)

    def __setitem__(self, key: str, value: str) -> None:
        self.set(key, value)

    def load(self):
        cp = ConfigParser(interpolation=None)
        cp.optionxform = str
        path = self.path
        if self.app and not path:
            path = self.app.get_settings_path()
        if not path:
            print("no settings path")
            return
        if os.path.exists(path):
            print("loading settings from", path)
        else:
            print("settings file", path, "does not exist")
        try:
            cp.read([path], encoding="UTF-8")
        except Exception as e:
            print(repr(e))
            return
        try:
            keys = cp.options("settings")
        except NoSectionError:
            return

        for key in keys:
            if key.startswith("__"):
                print("Ignoring setting", key)
                continue
            value = cp.get("settings", key)
            print(key, value)
            self.values[key] = value

    def save(self, extra: Dict[str, str]=None) -> None:
        partial_path = self.path + ".partial"
        print("writing settings to", partial_path)

        save_values = {}  # type: Dict[Tuple[str, str], str]
        for key, value in self.values.items():
            save_values[("settings", str(key))] = str(value)
        if extra is not None:
            for key, value in extra.items():
                try:
                    section, key = key.split("/")
                except ValueError:
                    section = "settings"
                save_values[(section, key)] = value

        cp = ConfigParser(interpolation=None)
        cp.optionxform = str
        # We want the settings section to be listed first.
        cp.add_section("settings")

        for (section, key) in sorted(save_values.keys()):
            value = save_values[(section, key)]
            if not value:
                # We do not need to write empty values, as non-existing
                # keys have an implicit empty value.
                continue
            if not cp.has_section(section):
                cp.add_section(section)
            cp.set(section, key, value)

        with open(partial_path, "w", encoding="UTF-8", newline="\n") as f:
            cp.write(f)
        print("moving settings file to", self.path)
        shutil.move(partial_path, self.path)


_path = ""
_settings = None  # type: Settings


def get(key: str) -> str:
    load()
    return _settings.get(key)


def set(key: str, value: str) -> None:
    load()
    _settings.set(key, value)


def load() -> None:
    global _settings
    if not _settings:
        # assert _path
        _settings = Settings(path=_path)


def save() -> None:
    assert _path
    _settings.save()


def unload() -> None:
    global _settings
    _settings = None


def set_path(path: str) -> None:
    global _path
    _path = path


def _log_key_value(key, value):
    if "username" in key or "password" in key or "auth" in key \
            or "email" in key:
        print("setting set {0} to *CENSORED*".format(key))
    else:
        print("settings set {0} to {1}".format(key, value))



from .common.element import Element, LightElement
from .common.group import Group
from .common.layout import VerticalLayout, HorizontalLayout
from .common.spacer import Spacer


default_window_parent = []
default_window_center = []

toolkit = "qt"
use_qt = True


from .qt import *
from .qt.Application import Application
from .qt.Button import Button
from .qt.CheckBox import CheckBox, HeadingCheckBox
from .qt.Color import Color
from .qt.Choice import Choice, ItemChoice
from .qt.ComboBox import ComboBox
from .qt.dialog import Dialog
from .qt.dialogbuttons import DialogButtons
from .qt.DirDialog import DirDialog
from .qt.FileDialog import FileDialog, pick_directory, pick_file, pick_files
from .qt.HeadingLabel import HeadingLabel
from .qt.Icon import Icon
from .qt.Image import Image
from .qt.ImageButton import ImageButton
from .qt.ImageView import ImageView
from .qt.label import Label, PlainLabel, URLLabel, MultiLineLabel
from .qt.legacydialog import LegacyDialog
from .qt.ListView import ListView
from .qt.Menu import Menu
from .qt.Panel import Panel
from .qt.scrollarea import VerticalScrollArea
from .qt.SpinCtrl import SpinCtrl
from .qt.System import System
from .qt.timer import IntervalTimer
from .qt.TextArea import TextArea
from .qt.TextField import TextField, PasswordField
from .qt.VerticalItemView import VerticalItemView
from .qt.window import Window



import fsui as fsui


class ScanProgressGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()
        # self.layout.padding_left = 10
        # self.layout.padding_top = 10
        # self.layout.padding_right = 10
        # self.layout.padding_bottom = 10

        # #image = fsui.Image("fs_uae_launcher:res/search_group.png")
        # #self.image_view = fsui.ImageView(self, image)
        # self.layout.add_spacer(20)
        # #self.layout.add(self.image_view, valign=0.0)
        # self.layout.add_spacer(48)
        # self.layout.add_spacer(20)

        self.layout2 = fsui.VerticalLayout()
        self.layout.add(self.layout2, fill=True, expand=True)

        self.title_label = fsui.HeadingLabel(self, "")
        self.layout2.add(self.title_label, fill=True)

        # self.layout2.add_spacer(10)
        # hor_layout = fsui.HorizontalLayout()
        # self.layout2.add(hor_layout)

        # self.scan_label = fsui.Label(self, _("Scan for:"))
        # hor_layout.add(self.scan_label)
        # hor_layout.add_spacer(10)

        # self.scan_roms = fsui.CheckBox(self, _("ROMs"))
        # if Settings.get("scan_roms") == "1":
        #     self.scan_roms.check()
        # self.scan_roms.on_change = self.on_change
        # hor_layout.add(self.scan_roms)
        # hor_layout.add_spacer(10)

        # self.scan_files = fsui.CheckBox(self, _("Game Files"))
        # if Settings.get("scan_files") == "1":
        #     self.scan_files.check()
        # self.scan_files.on_change = self.on_change
        # hor_layout.add(self.scan_files)
        # hor_layout.add_spacer(10)

        # self.scan_configs = fsui.CheckBox(self, _("Configurations"))
        # if Settings.get("scan_configs") == "1":
        #     self.scan_configs.check()
        # self.scan_configs.on_change = self.on_change
        # hor_layout.add(self.scan_configs)
        # hor_layout.add_spacer(10)

        self.layout2.add_spacer(10)
        self.status_label = fsui.Label(self, "")
        self.layout2.add(self.status_label, fill=True)

    # def on_change(self):
    #     value = "1" if self.scan_roms.is_checked() else "0"
    #     Settings.set("scan_roms", value)
    #     value = "1" if self.scan_files.is_checked() else "0"
    #     Settings.set("scan_files", value)
    #     value = "1" if self.scan_configs.is_checked() else "0"
    #     Settings.set("scan_configs", value)



from fsgs.platform import PlatformHandler
from fsgs.mess.atari_2600 import Atari2600Runner
from .loader import SimpleLoader


class Atari2600PlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Atari 2600"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return Atari2600Loader(fsgs)

    def get_runner(self, fsgs):
        return Atari2600Runner(fsgs)


class Atari2600Loader(SimpleLoader):
    pass



from fsui.qt import QSignal, QObject, QX11Info
# print("a")
# try:
#     # noinspection PyUnresolvedReferences
#     from gi.repository import Gdk
# except ImportError:
#     Gdk = None
# print("b")
try:
    # noinspection PyUnresolvedReferences, PyPackageRequirements
    from gi import require_version
    require_version("Wnck", "1.0")
    # noinspection PyPackageRequirements
    from gi.repository import Wnck

    # require_version("Gtk", "2.0")
    # from gi.repository import Gtk

    # Wnck.enable_wnck(version="1.0")
    # print(dir(Wnck))
except ImportError:
    Wnck = None
print("c")


class WindowProxy:

    def __init__(self, window, changed=""):
        self.window = window
        self.changed = changed

    @property
    def name(self):
        if self.window is None:
            return ""
        return self.window.get_name()

    def activate(self):
        if self.window is None:
            return
        if self.window.is_active():
            self.window.minimize()
            return
        # self.window.activate(time.time())
        # noinspection PyArgumentList
        self.window.activate(QX11Info.appTime())

    def is_active(self):
        if self.window is None:
            return False
        return self.window.is_active()

    def is_minimized(self):
        if self.window is None:
            return False
        return self.window.is_minimized()

    def is_desktop(self):
        if self.window is None:
            return False
        return self.window.get_window_type() == Wnck.WindowType.DESKTOP

    def open_menu(self, bottom_left):
        # menu = Wnck.ActionMenu.new(self.window)
        # print(menu)
        # #Gtk.Menu.menu_popup(menu)
        # print(dir(menu))
        # menu.menu_popup()
        pass

    def __eq__(self, other):
        return self.window == other.window


class RealWindowHandler(QObject):

    window_added = QSignal(WindowProxy)
    window_removed = QSignal(WindowProxy)
    window_activated = QSignal(WindowProxy)
    window_changed = QSignal(WindowProxy)

    def __init__(self):
        QObject.__init__(self)
        if Wnck is None:
            return
        self.screen = Wnck.Screen.get_default()
        # self.screen.force_update()
        print(self.screen)
        self.screen.connect("window-opened", self.__window_opened)
        self.screen.connect("window-closed", self.__window_closed)
        self.screen.connect("active-workspace-changed",
                            self.__active_workspace_changed)
        self.screen.connect("active-window-changed",
                            self.__active_window_changed)

        self.workspace_index = 0
        self.workspace = None
        # workspace = self.screen.get_workspace()
        # self.workspace_index = self.screen.get_workspace()
        self.all_windows = []
        self.workspace_windows = []
        self.active_window = None

    def __active_window_changed(self, screen, old_window):
        print("__active_window_changed", old_window)
        if screen != self.screen:
            return
        self.active_window = screen.get_active_window()
        print("emit", self.active_window)
        self.window_activated.emit(WindowProxy(self.active_window))
        self.window_changed.emit(WindowProxy(old_window, "active"))
        self.window_changed.emit(WindowProxy(self.active_window, "active"))

    def __window_opened(self, screen, window):
        if screen != self.screen:
            return
        print(screen, window)
        self.all_windows.append(window)
        if self.is_window_relevant(window):
            self.add_workspace_window(window)

    def __window_closed(self, screen, window):
        if screen != self.screen:
            return
        if window in self.workspace_windows:
            self.remove_workspace_window(window)
        self.all_windows.remove(window)

    def is_window_relevant(self, window):
        if window.is_skip_tasklist():
            return False
        workspace = window.get_workspace()
        if workspace is not None and workspace != self.workspace:
            return False
        #    # pinned to all workspaces
        #    return True
        # if workspace.get_number() != self.workspace_index:
        #     return False
        # if workspace != self.workspace:
        #     return False
        return True

    def add_workspace_window(self, window):
        print("- add", window)
        print(" ", window.get_name())
        self.workspace_windows.append(window)

        window.connect("state-changed", self.__window_state_changed)
        window.connect("name-changed", self.__window_name_changed)

        print("emitting window added signal")
        self.window_added.emit(WindowProxy(window))

    def remove_workspace_window(self, window):
        print("- remove", window)
        # window.disconnect(self.__window_state_changed)
        self.workspace_windows.remove(window)
        self.window_removed.emit(WindowProxy(window))

    def __window_state_changed(self, window, changed_mask, new_state):
        self.window_changed.emit(WindowProxy(window, "state"))

    def __window_name_changed(self, window):
        print("__window_name_changed", window)
        self.window_changed.emit(WindowProxy(window, "name"))

    def __active_workspace_changed(self, screen, previous_workspace):
        if screen != self.screen:
            return
        print("new workspace", screen, previous_workspace)
        self.workspace = self.screen.get_active_workspace()
        self.workspace_index = self.workspace.get_number()

        print("workspace index is", self.workspace_index)
        add_windows = []
        remove_windows = []
        for window in self.all_windows:
            if self.is_window_relevant(window):
                if window not in self.workspace_windows:
                    add_windows.append(window)
            else:
                if window in self.workspace_windows:
                    remove_windows.append(window)

        for window in remove_windows:
            self.remove_workspace_window(window)
        for window in add_windows:
            self.add_workspace_window(window)



from fsgs.ogd.client import OGDClient
from fsui import Image
from ...Config import Config
from .WebButton import WebButton


class EditButton(WebButton):

    def __init__(self, parent):
        icon = Image("fs_uae_launcher:res/16/pencil.png")
        WebButton.__init__(self, parent, icon)

    def get_url(self):
        variant_uuid = Config.get("variant_uuid", "")
        if not variant_uuid:
            return
        return "http://{0}/game/{1}/edit#{1}".format(
            OGDClient.get_server(), variant_uuid)



import json
from collections import defaultdict
from fsgs.ogd.client import OGDClient


class SimpleLoader:

    def __init__(self, fsgs):
        self.fsgs = fsgs
        self.config = {}

    def get_config(self):
        return self.config.copy()

    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        assert len(file_list) == 1
        self.config["cartridge"] = "sha1://{0}/{1}".format(
            file_list[0]["sha1"], file_list[0]["name"])

    def load_extra(self, values):
        pass

    def load_basic(self, values):
        for key in ["command"]:
            if key in values:
                self.config[key] = values[key]

        self.config["platform"] = values["platform"]
        self.config["model"] = values["model"]

        self.config["viewport"] = values["viewport"]
        self.config["video_standard"] = values["video_standard"]

    def load_info(self, values):
        self.config["languages"] = values["languages"]
        self.config["players"] = values["players"]
        self.config["year"] = values["year"]
        self.config["publisher"] = values["publisher"]
        self.config["developer"] = values["developer"]

        self.config["x_game_notice"] = values["game_notice"]
        self.config["x_variant_notice"] = values["variant_notice"]
        self.config["x_variant_warning"] = values["variant_warning"]
        self.config["x_variant_error"] = values["variant_error"]

        self.config["database_url"] = "http://{0}/game/{1}".format(
            OGDClient.get_server(), values["parent_uuid"])
        for key in ["mobygames_url"]:
            self.config[key] = values[key]

        for key in ["download_file", "download_page", "download_terms",
                    "download_notice"]:
            if key in values:
                self.config[key] = values[key]

    def load_images(self, values):
        for key in ["front_sha1", "screen1_sha1", "screen2_sha1",
                    "screen3_sha1", "screen4_sha1", "screen5_sha1",
                    "title_sha1"]:
            self.config[key] = values[key]

    def load_values(self, key_values):
        # print(key_values)
        values = defaultdict(str)
        values.update(key_values)

        self.load_basic(values)
        self.load_files(values)
        self.load_extra(values)
        self.load_info(values)
        self.load_images(values)

        return self.get_config()



'''OpenGL extension VERSION.GL_1_1

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_1_1 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_1_1.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_1_1 import *
from OpenGL.raw.GL.VERSION.GL_1_1 import _EXTENSION_NAME

def glInitGl11VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
from OpenGL.raw.GL.VERSION.GL_1_0 import *
GL_MODELVIEW0_EXT = GL_MODELVIEW # alias
GL_MODELVIEW0_MATRIX_EXT = GL_MODELVIEW_MATRIX # alias
GL_MODELVIEW0_STACK_DEPTH_EXT = GL_MODELVIEW_STACK_DEPTH # alias
GL_TEXTURE_COMPONENTS = GL_TEXTURE_INTERNAL_FORMAT # alias

glAreTexturesResident = arrays.setInputArraySizeType(
    glAreTexturesResident,
    None, # XXX Could not determine size of argument textures for glAreTexturesResident arrays.GLuintArray
    arrays.GLuintArray, 
    'textures',
)

glBitmap = arrays.setInputArraySizeType(
    glBitmap,
    None, # XXX Could not determine size of argument bitmap for glBitmap arrays.GLubyteArray
    arrays.GLubyteArray, 
    'bitmap',
)

glClipPlane = arrays.setInputArraySizeType(
    glClipPlane,
    None, # XXX Could not determine size of argument equation for glClipPlane arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'equation',
)

glColor3bv = arrays.setInputArraySizeType(
    glColor3bv,
    3,
    arrays.GLbyteArray,
    'v',
)

glColor3dv = arrays.setInputArraySizeType(
    glColor3dv,
    3,
    arrays.GLdoubleArray,
    'v',
)

glColor3fv = arrays.setInputArraySizeType(
    glColor3fv,
    3,
    arrays.GLfloatArray,
    'v',
)

glColor3iv = arrays.setInputArraySizeType(
    glColor3iv,
    3,
    arrays.GLintArray,
    'v',
)

glColor3sv = arrays.setInputArraySizeType(
    glColor3sv,
    3,
    arrays.GLshortArray,
    'v',
)

glColor3ubv = arrays.setInputArraySizeType(
    glColor3ubv,
    3,
    arrays.GLubyteArray,
    'v',
)

glColor3uiv = arrays.setInputArraySizeType(
    glColor3uiv,
    3,
    arrays.GLuintArray,
    'v',
)

glColor3usv = arrays.setInputArraySizeType(
    glColor3usv,
    3,
    arrays.GLushortArray,
    'v',
)

glColor4bv = arrays.setInputArraySizeType(
    glColor4bv,
    4,
    arrays.GLbyteArray,
    'v',
)

glColor4dv = arrays.setInputArraySizeType(
    glColor4dv,
    4,
    arrays.GLdoubleArray,
    'v',
)

glColor4fv = arrays.setInputArraySizeType(
    glColor4fv,
    4,
    arrays.GLfloatArray,
    'v',
)

glColor4iv = arrays.setInputArraySizeType(
    glColor4iv,
    4,
    arrays.GLintArray,
    'v',
)

glColor4sv = arrays.setInputArraySizeType(
    glColor4sv,
    4,
    arrays.GLshortArray,
    'v',
)

glColor4ubv = arrays.setInputArraySizeType(
    glColor4ubv,
    4,
    arrays.GLubyteArray,
    'v',
)

glColor4uiv = arrays.setInputArraySizeType(
    glColor4uiv,
    4,
    arrays.GLuintArray,
    'v',
)

glColor4usv = arrays.setInputArraySizeType(
    glColor4usv,
    4,
    arrays.GLushortArray,
    'v',
)

glDeleteTextures = arrays.setInputArraySizeType(
    glDeleteTextures,
    None, # XXX Could not determine size of argument textures for glDeleteTextures arrays.GLuintArray
    arrays.GLuintArray, 
    'textures',
)

glEvalCoord1dv = arrays.setInputArraySizeType(
    glEvalCoord1dv,
    1,
    arrays.GLdoubleArray,
    'u',
)

glEvalCoord1fv = arrays.setInputArraySizeType(
    glEvalCoord1fv,
    1,
    arrays.GLfloatArray,
    'u',
)

glEvalCoord2dv = arrays.setInputArraySizeType(
    glEvalCoord2dv,
    2,
    arrays.GLdoubleArray,
    'u',
)

glEvalCoord2fv = arrays.setInputArraySizeType(
    glEvalCoord2fv,
    2,
    arrays.GLfloatArray,
    'u',
)

glFeedbackBuffer = arrays.setInputArraySizeType(
    glFeedbackBuffer,
    None, # XXX Could not determine size of argument buffer for glFeedbackBuffer arrays.GLfloatArray
    arrays.GLfloatArray, 
    'buffer',
)

glFogfv = arrays.setInputArraySizeType(
    glFogfv,
    None, # XXX Could not determine size of argument params for glFogfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glFogiv = arrays.setInputArraySizeType(
    glFogiv,
    None, # XXX Could not determine size of argument params for glFogiv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glGenTextures = arrays.setInputArraySizeType(
    glGenTextures,
    None, # XXX Could not determine size of argument textures for glGenTextures arrays.GLuintArray
    arrays.GLuintArray, 
    'textures',
)

glGetClipPlane = arrays.setInputArraySizeType(
    glGetClipPlane,
    None, # XXX Could not determine size of argument equation for glGetClipPlane arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'equation',
)

glGetDoublev = arrays.setInputArraySizeType(
    glGetDoublev,
    None, # XXX Could not determine size of argument params for glGetDoublev arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'data',
)

glGetFloatv = arrays.setInputArraySizeType(
    glGetFloatv,
    None, # XXX Could not determine size of argument params for glGetFloatv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'data',
)

glGetIntegerv = arrays.setInputArraySizeType(
    glGetIntegerv,
    None, # XXX Could not determine size of argument params for glGetIntegerv arrays.GLintArray
    arrays.GLintArray, 
    'data',
)

glGetLightfv = arrays.setInputArraySizeType(
    glGetLightfv,
    None, # XXX Could not determine size of argument params for glGetLightfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glGetLightiv = arrays.setInputArraySizeType(
    glGetLightiv,
    None, # XXX Could not determine size of argument params for glGetLightiv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glGetMapdv = arrays.setInputArraySizeType(
    glGetMapdv,
    None, # XXX Could not determine size of argument v for glGetMapdv arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'v',
)

glGetMapfv = arrays.setInputArraySizeType(
    glGetMapfv,
    None, # XXX Could not determine size of argument v for glGetMapfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'v',
)

glGetMapiv = arrays.setInputArraySizeType(
    glGetMapiv,
    None, # XXX Could not determine size of argument v for glGetMapiv arrays.GLintArray
    arrays.GLintArray, 
    'v',
)

glGetMaterialfv = arrays.setInputArraySizeType(
    glGetMaterialfv,
    None, # XXX Could not determine size of argument params for glGetMaterialfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glGetMaterialiv = arrays.setInputArraySizeType(
    glGetMaterialiv,
    None, # XXX Could not determine size of argument params for glGetMaterialiv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glGetPixelMapfv = arrays.setInputArraySizeType(
    glGetPixelMapfv,
    None, # XXX Could not determine size of argument values for glGetPixelMapfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'values',
)

glGetPixelMapuiv = arrays.setInputArraySizeType(
    glGetPixelMapuiv,
    None, # XXX Could not determine size of argument values for glGetPixelMapuiv arrays.GLuintArray
    arrays.GLuintArray, 
    'values',
)

glGetPixelMapusv = arrays.setInputArraySizeType(
    glGetPixelMapusv,
    None, # XXX Could not determine size of argument values for glGetPixelMapusv arrays.GLushortArray
    arrays.GLushortArray, 
    'values',
)

glGetPolygonStipple = arrays.setInputArraySizeType(
    glGetPolygonStipple,
    None, # XXX Could not determine size of argument mask for glGetPolygonStipple arrays.GLubyteArray
    arrays.GLubyteArray, 
    'mask',
)

glGetTexEnvfv = arrays.setInputArraySizeType(
    glGetTexEnvfv,
    None, # XXX Could not determine size of argument params for glGetTexEnvfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glGetTexEnviv = arrays.setInputArraySizeType(
    glGetTexEnviv,
    None, # XXX Could not determine size of argument params for glGetTexEnviv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glGetTexGendv = arrays.setInputArraySizeType(
    glGetTexGendv,
    None, # XXX Could not determine size of argument params for glGetTexGendv arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'params',
)

glGetTexGenfv = arrays.setInputArraySizeType(
    glGetTexGenfv,
    None, # XXX Could not determine size of argument params for glGetTexGenfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glGetTexGeniv = arrays.setInputArraySizeType(
    glGetTexGeniv,
    None, # XXX Could not determine size of argument params for glGetTexGeniv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glGetTexLevelParameterfv = arrays.setInputArraySizeType(
    glGetTexLevelParameterfv,
    None, # XXX Could not determine size of argument params for glGetTexLevelParameterfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glGetTexLevelParameteriv = arrays.setInputArraySizeType(
    glGetTexLevelParameteriv,
    None, # XXX Could not determine size of argument params for glGetTexLevelParameteriv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glGetTexParameterfv = arrays.setInputArraySizeType(
    glGetTexParameterfv,
    None, # XXX Could not determine size of argument params for glGetTexParameterfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glGetTexParameteriv = arrays.setInputArraySizeType(
    glGetTexParameteriv,
    None, # XXX Could not determine size of argument params for glGetTexParameteriv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glIndexdv = arrays.setInputArraySizeType(
    glIndexdv,
    None, # XXX Could not determine size of argument c for glIndexdv arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'c',
)

glIndexfv = arrays.setInputArraySizeType(
    glIndexfv,
    None, # XXX Could not determine size of argument c for glIndexfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'c',
)

glIndexiv = arrays.setInputArraySizeType(
    glIndexiv,
    None, # XXX Could not determine size of argument c for glIndexiv arrays.GLintArray
    arrays.GLintArray, 
    'c',
)

glIndexsv = arrays.setInputArraySizeType(
    glIndexsv,
    None, # XXX Could not determine size of argument c for glIndexsv arrays.GLshortArray
    arrays.GLshortArray, 
    'c',
)

glIndexubv = arrays.setInputArraySizeType(
    glIndexubv,
    None, # XXX Could not determine size of argument c for glIndexubv arrays.GLubyteArray
    arrays.GLubyteArray, 
    'c',
)

glLightModelfv = arrays.setInputArraySizeType(
    glLightModelfv,
    None, # XXX Could not determine size of argument params for glLightModelfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glLightModeliv = arrays.setInputArraySizeType(
    glLightModeliv,
    None, # XXX Could not determine size of argument params for glLightModeliv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glLightfv = arrays.setInputArraySizeType(
    glLightfv,
    None, # XXX Could not determine size of argument params for glLightfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glLightiv = arrays.setInputArraySizeType(
    glLightiv,
    None, # XXX Could not determine size of argument params for glLightiv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glLoadMatrixd = arrays.setInputArraySizeType(
    glLoadMatrixd,
    16,
    arrays.GLdoubleArray,
    'm',
)

glLoadMatrixf = arrays.setInputArraySizeType(
    glLoadMatrixf,
    16,
    arrays.GLfloatArray,
    'm',
)

glMap1d = arrays.setInputArraySizeType(
    glMap1d,
    None, # XXX Could not determine size of argument points for glMap1d arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'points',
)

glMap1f = arrays.setInputArraySizeType(
    glMap1f,
    None, # XXX Could not determine size of argument points for glMap1f arrays.GLfloatArray
    arrays.GLfloatArray, 
    'points',
)

glMap2d = arrays.setInputArraySizeType(
    glMap2d,
    None, # XXX Could not determine size of argument points for glMap2d arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'points',
)

glMap2f = arrays.setInputArraySizeType(
    glMap2f,
    None, # XXX Could not determine size of argument points for glMap2f arrays.GLfloatArray
    arrays.GLfloatArray, 
    'points',
)

glMaterialfv = arrays.setInputArraySizeType(
    glMaterialfv,
    None, # XXX Could not determine size of argument params for glMaterialfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glMaterialiv = arrays.setInputArraySizeType(
    glMaterialiv,
    None, # XXX Could not determine size of argument params for glMaterialiv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glMultMatrixd = arrays.setInputArraySizeType(
    glMultMatrixd,
    16,
    arrays.GLdoubleArray,
    'm',
)

glMultMatrixf = arrays.setInputArraySizeType(
    glMultMatrixf,
    16,
    arrays.GLfloatArray,
    'm',
)

glNormal3bv = arrays.setInputArraySizeType(
    glNormal3bv,
    3,
    arrays.GLbyteArray,
    'v',
)

glNormal3dv = arrays.setInputArraySizeType(
    glNormal3dv,
    3,
    arrays.GLdoubleArray,
    'v',
)

glNormal3fv = arrays.setInputArraySizeType(
    glNormal3fv,
    3,
    arrays.GLfloatArray,
    'v',
)

glNormal3iv = arrays.setInputArraySizeType(
    glNormal3iv,
    3,
    arrays.GLintArray,
    'v',
)

glNormal3sv = arrays.setInputArraySizeType(
    glNormal3sv,
    3,
    arrays.GLshortArray,
    'v',
)

glPixelMapfv = arrays.setInputArraySizeType(
    glPixelMapfv,
    None, # XXX Could not determine size of argument values for glPixelMapfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'values',
)

glPixelMapuiv = arrays.setInputArraySizeType(
    glPixelMapuiv,
    None, # XXX Could not determine size of argument values for glPixelMapuiv arrays.GLuintArray
    arrays.GLuintArray, 
    'values',
)

glPixelMapusv = arrays.setInputArraySizeType(
    glPixelMapusv,
    None, # XXX Could not determine size of argument values for glPixelMapusv arrays.GLushortArray
    arrays.GLushortArray, 
    'values',
)

glPolygonStipple = arrays.setInputArraySizeType(
    glPolygonStipple,
    None, # XXX Could not determine size of argument mask for glPolygonStipple arrays.GLubyteArray
    arrays.GLubyteArray, 
    'mask',
)

glPrioritizeTextures = arrays.setInputArraySizeType(
    arrays.setInputArraySizeType(
        glPrioritizeTextures,
        None, # XXX Could not determine size of argument textures for glPrioritizeTextures arrays.GLuintArray
        arrays.GLuintArray, 
        'textures',
    ),
    None, # XXX Could not determine size of argument priorities for glPrioritizeTextures arrays.GLclampfArray
    arrays.GLclampfArray, 
    'priorities',
)

glRasterPos2dv = arrays.setInputArraySizeType(
    glRasterPos2dv,
    2,
    arrays.GLdoubleArray,
    'v',
)

glRasterPos2fv = arrays.setInputArraySizeType(
    glRasterPos2fv,
    2,
    arrays.GLfloatArray,
    'v',
)

glRasterPos2iv = arrays.setInputArraySizeType(
    glRasterPos2iv,
    2,
    arrays.GLintArray,
    'v',
)

glRasterPos2sv = arrays.setInputArraySizeType(
    glRasterPos2sv,
    2,
    arrays.GLshortArray,
    'v',
)

glRasterPos3dv = arrays.setInputArraySizeType(
    glRasterPos3dv,
    3,
    arrays.GLdoubleArray,
    'v',
)

glRasterPos3fv = arrays.setInputArraySizeType(
    glRasterPos3fv,
    3,
    arrays.GLfloatArray,
    'v',
)

glRasterPos3iv = arrays.setInputArraySizeType(
    glRasterPos3iv,
    3,
    arrays.GLintArray,
    'v',
)

glRasterPos3sv = arrays.setInputArraySizeType(
    glRasterPos3sv,
    3,
    arrays.GLshortArray,
    'v',
)

glRasterPos4dv = arrays.setInputArraySizeType(
    glRasterPos4dv,
    4,
    arrays.GLdoubleArray,
    'v',
)

glRasterPos4fv = arrays.setInputArraySizeType(
    glRasterPos4fv,
    4,
    arrays.GLfloatArray,
    'v',
)

glRasterPos4iv = arrays.setInputArraySizeType(
    glRasterPos4iv,
    4,
    arrays.GLintArray,
    'v',
)

glRasterPos4sv = arrays.setInputArraySizeType(
    glRasterPos4sv,
    4,
    arrays.GLshortArray,
    'v',
)

glRectdv = arrays.setInputArraySizeType(
    arrays.setInputArraySizeType(
        glRectdv,
        None, # XXX Could not determine size of argument v1 for glRectdv arrays.GLdoubleArray
        arrays.GLdoubleArray, 
        'v1',
    ),
    None, # XXX Could not determine size of argument v2 for glRectdv arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'v2',
)

glRectfv = arrays.setInputArraySizeType(
    arrays.setInputArraySizeType(
        glRectfv,
        None, # XXX Could not determine size of argument v1 for glRectfv arrays.GLfloatArray
        arrays.GLfloatArray, 
        'v1',
    ),
    None, # XXX Could not determine size of argument v2 for glRectfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'v2',
)

glRectiv = arrays.setInputArraySizeType(
    arrays.setInputArraySizeType(
        glRectiv,
        None, # XXX Could not determine size of argument v1 for glRectiv arrays.GLintArray
        arrays.GLintArray, 
        'v1',
    ),
    None, # XXX Could not determine size of argument v2 for glRectiv arrays.GLintArray
    arrays.GLintArray, 
    'v2',
)

glRectsv = arrays.setInputArraySizeType(
    arrays.setInputArraySizeType(
        glRectsv,
        None, # XXX Could not determine size of argument v1 for glRectsv arrays.GLshortArray
        arrays.GLshortArray, 
        'v1',
    ),
    None, # XXX Could not determine size of argument v2 for glRectsv arrays.GLshortArray
    arrays.GLshortArray, 
    'v2',
)

glSelectBuffer = arrays.setInputArraySizeType(
    glSelectBuffer,
    None, # XXX Could not determine size of argument buffer for glSelectBuffer arrays.GLuintArray
    arrays.GLuintArray, 
    'buffer',
)

glTexCoord1dv = arrays.setInputArraySizeType(
    glTexCoord1dv,
    1,
    arrays.GLdoubleArray,
    'v',
)

glTexCoord1fv = arrays.setInputArraySizeType(
    glTexCoord1fv,
    1,
    arrays.GLfloatArray,
    'v',
)

glTexCoord1iv = arrays.setInputArraySizeType(
    glTexCoord1iv,
    1,
    arrays.GLintArray,
    'v',
)

glTexCoord1sv = arrays.setInputArraySizeType(
    glTexCoord1sv,
    1,
    arrays.GLshortArray,
    'v',
)

glTexCoord2dv = arrays.setInputArraySizeType(
    glTexCoord2dv,
    2,
    arrays.GLdoubleArray,
    'v',
)

glTexCoord2fv = arrays.setInputArraySizeType(
    glTexCoord2fv,
    2,
    arrays.GLfloatArray,
    'v',
)

glTexCoord2iv = arrays.setInputArraySizeType(
    glTexCoord2iv,
    2,
    arrays.GLintArray,
    'v',
)

glTexCoord2sv = arrays.setInputArraySizeType(
    glTexCoord2sv,
    2,
    arrays.GLshortArray,
    'v',
)

glTexCoord3dv = arrays.setInputArraySizeType(
    glTexCoord3dv,
    3,
    arrays.GLdoubleArray,
    'v',
)

glTexCoord3fv = arrays.setInputArraySizeType(
    glTexCoord3fv,
    3,
    arrays.GLfloatArray,
    'v',
)

glTexCoord3iv = arrays.setInputArraySizeType(
    glTexCoord3iv,
    3,
    arrays.GLintArray,
    'v',
)

glTexCoord3sv = arrays.setInputArraySizeType(
    glTexCoord3sv,
    3,
    arrays.GLshortArray,
    'v',
)

glTexCoord4dv = arrays.setInputArraySizeType(
    glTexCoord4dv,
    4,
    arrays.GLdoubleArray,
    'v',
)

glTexCoord4fv = arrays.setInputArraySizeType(
    glTexCoord4fv,
    4,
    arrays.GLfloatArray,
    'v',
)

glTexCoord4iv = arrays.setInputArraySizeType(
    glTexCoord4iv,
    4,
    arrays.GLintArray,
    'v',
)

glTexCoord4sv = arrays.setInputArraySizeType(
    glTexCoord4sv,
    4,
    arrays.GLshortArray,
    'v',
)

glTexEnvfv = arrays.setInputArraySizeType(
    glTexEnvfv,
    None, # XXX Could not determine size of argument params for glTexEnvfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glTexEnviv = arrays.setInputArraySizeType(
    glTexEnviv,
    None, # XXX Could not determine size of argument params for glTexEnviv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glTexGendv = arrays.setInputArraySizeType(
    glTexGendv,
    None, # XXX Could not determine size of argument params for glTexGendv arrays.GLdoubleArray
    arrays.GLdoubleArray, 
    'params',
)

glTexGenfv = arrays.setInputArraySizeType(
    glTexGenfv,
    None, # XXX Could not determine size of argument params for glTexGenfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glTexGeniv = arrays.setInputArraySizeType(
    glTexGeniv,
    None, # XXX Could not determine size of argument params for glTexGeniv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glTexParameterfv = arrays.setInputArraySizeType(
    glTexParameterfv,
    None, # XXX Could not determine size of argument params for glTexParameterfv arrays.GLfloatArray
    arrays.GLfloatArray, 
    'params',
)

glTexParameteriv = arrays.setInputArraySizeType(
    glTexParameteriv,
    None, # XXX Could not determine size of argument params for glTexParameteriv arrays.GLintArray
    arrays.GLintArray, 
    'params',
)

glVertex2dv = arrays.setInputArraySizeType(
    glVertex2dv,
    2,
    arrays.GLdoubleArray,
    'v',
)

glVertex2fv = arrays.setInputArraySizeType(
    glVertex2fv,
    2,
    arrays.GLfloatArray,
    'v',
)

glVertex2iv = arrays.setInputArraySizeType(
    glVertex2iv,
    2,
    arrays.GLintArray,
    'v',
)

glVertex2sv = arrays.setInputArraySizeType(
    glVertex2sv,
    2,
    arrays.GLshortArray,
    'v',
)

glVertex3dv = arrays.setInputArraySizeType(
    glVertex3dv,
    3,
    arrays.GLdoubleArray,
    'v',
)

glVertex3fv = arrays.setInputArraySizeType(
    glVertex3fv,
    3,
    arrays.GLfloatArray,
    'v',
)

glVertex3iv = arrays.setInputArraySizeType(
    glVertex3iv,
    3,
    arrays.GLintArray,
    'v',
)

glVertex3sv = arrays.setInputArraySizeType(
    glVertex3sv,
    3,
    arrays.GLshortArray,
    'v',
)

glVertex4dv = arrays.setInputArraySizeType(
    glVertex4dv,
    4,
    arrays.GLdoubleArray,
    'v',
)

glVertex4fv = arrays.setInputArraySizeType(
    glVertex4fv,
    4,
    arrays.GLfloatArray,
    'v',
)

glVertex4iv = arrays.setInputArraySizeType(
    glVertex4iv,
    4,
    arrays.GLintArray,
    'v',
)

glVertex4sv = arrays.setInputArraySizeType(
    glVertex4sv,
    4,
    arrays.GLshortArray,
    'v',
)



import fsui as fsui
from fsui.extra.iconheader import NewIconHeader
from ...I18N import gettext
from .OptionUI import OptionUI


class VideoSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("video-settings", "pkg:fs_uae_workspace"),
            gettext("Video Settings"),
            "")
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        def add_option(name):
            self.layout.add(OptionUI.create_group(self, name), fill=True,
                            margin_top=10, margin_bottom=10)

        add_option("fullscreen")
        add_option("monitor")

        label = fsui.HeadingLabel(self, gettext("Scaling"))
        self.layout.add(label, margin_top=20, margin_bottom=20)

        add_option("zoom")
        add_option("keep_aspect")

        label = fsui.HeadingLabel(self, gettext("Filters"))
        self.layout.add(label, margin_top=20, margin_bottom=20)

        add_option("scanlines")
        add_option("rtg_scanlines")

        # add_option("video_format")
        #
        # add_option("video_sync")
        # add_option("low_latency_vsync")
        #
        # label = fsui.HeadingLabel(self, gettext("OpenGL Settings"))
        # self.layout.add(label, margin_top=20, margin_bottom=20)
        #
        # add_option("fsaa")
        # add_option("texture_filter")
        # add_option("video_sync_method")
        # add_option("texture_format")



'''OpenGL extension VERSION.GL_4_0

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_4_0 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_4_0.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_4_0 import *
from OpenGL.raw.GL.VERSION.GL_4_0 import _EXTENSION_NAME

def glInitGl40VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
from OpenGL.GL.ARB.texture_query_lod import *
from OpenGL.GL.ARB.draw_indirect import *
from OpenGL.GL.ARB.gpu_shader5 import *
from OpenGL.GL.ARB.gpu_shader_fp64 import *
from OpenGL.GL.ARB.shader_subroutine import *
from OpenGL.GL.ARB.tessellation_shader import *
from OpenGL.GL.ARB.texture_buffer_object_rgb32 import *
from OpenGL.GL.ARB.texture_cube_map_array import *
from OpenGL.GL.ARB.texture_gather import *
from OpenGL.GL.ARB.transform_feedback2 import *
from OpenGL.GL.ARB.transform_feedback3 import *



from fsgs.Database import Database
from fsui import ImageButton, Image
from fsgs.OAGDClient import OAGDClient
from ...Config import Config
from ...I18N import gettext
from ..IconButton import IconButton


class RatingButton(ImageButton):

    def __init__(self, parent, rating):
        self.rating = rating
        self.icons = [None, None]
        if self.rating == 5:
            self.icons[0] = Image(
                "fs_uae_launcher:res/16/rating_fav_2.png")
            tooltip = gettext("Mark This Game Variant as Personal Favorite")
        elif self.rating == 4:
            self.icons[0] = Image(
                "fs_uae_launcher:res/16/thumb_up_2.png")
            tooltip = gettext("Like This Game Variant")
        elif self.rating == 1:
            self.icons[0] = Image(
                "fs_uae_launcher:res/16/thumb_down_2.png")
            tooltip = gettext("Dislike This Game Variant")
        else:
            raise Exception("ImageButton: unknown rating")
        self.icons[1] = self.icons[0].grey_scale()
        self.active_icon = 1
        ImageButton.__init__(self, parent, self.icons[1])
        self.set_min_width(IconButton.BUTTON_WIDTH)
        self.set_tooltip(tooltip)
        Config.add_listener(self)
        self.on_config("variant_rating", "")

    def on_destroy(self):
        Config.remove_listener(self)
        pass

    def on_activate(self):
        variant_uuid = Config.get("variant_uuid", "")
        if not variant_uuid:
            return
        client = OAGDClient()
        variant_rating = Config.get("variant_rating", "")
        if self.is_active(variant_rating):
            like = 0
        else:
            like = self.rating
        client.rate_variant(variant_uuid, like=like)

        like_rating = client.get("like", 0)
        work_rating = client.get("work", 0)
        database = Database.instance()
        cursor = database.cursor()
        # cursor.execute("UPDATE rating SET like_rating = ?, "
        #                "work_rating = ? WHERE game_uuid = ?",
        #                (like_rating, work_rating, variant_uuid))

        cursor.execute(
            "DELETE FROM rating WHERE game_uuid = ?", (variant_uuid,))
        cursor.execute(
            "INSERT INTO rating (game_uuid, work_rating, like_rating) "
            "VALUES (?, ?, ?)", (variant_uuid, work_rating, like_rating))
        database.commit()
        Config.set("variant_rating", str(like_rating))

    def is_active(self, value):
        try:
            rating = int(value)
        except ValueError:
            rating = 0
        if self.rating == 5:
            if rating == 5:
                return 1
            else:
                return 0
        elif self.rating == 4:
            if 3 <= rating <= 4:
                return 1
            else:
                return 0
        elif self.rating == 1:
            if 1 <= rating <= 2:
                return 1
            else:
                return 0

    def on_config(self, key, value):
        if key == "variant_rating":
            active_icon = not self.is_active(value)
            if active_icon != self.active_icon:
                self.set_image(self.icons[active_icon])
                self.active_icon = active_icon
                self.refresh()
        elif key == "variant_uuid":
            if value:
                if not self.is_enabled():
                    self.enable()
            else:
                if self.is_enabled():
                    self.disable()



import sys
import re
import traceback
import subprocess
from .I18N import gettext
from .Settings import Settings
from .Signal import Signal
from fsgs.amiga.FSUAEDeviceHelper import FSUAEDeviceHelper


def create_cmp_id(id):
    return id.lower().replace(" ", "")


class Device:

    def __init__(self, id, name, type):
        self.id = id
        self.name = name
        self.type = type
        self.port = None
        self.cmp_id = create_cmp_id(id)


class DeviceManager:

    initialized = False
    devices = []
    device_ids = []
    device_names = []
    device_types = []
    device_name_count = {}

    joystick_data = {}

    @classmethod
    def joystick_buttons(cls, device):
        return cls.joystick_data[device][0]

    @classmethod
    def joystick_hats(cls, device):
        return cls.joystick_data[device][1]

    @classmethod
    def joystick_axes(cls, device):
        return cls.joystick_data[device][2]

    @classmethod
    def joystick_balls(cls, device):
        return cls.joystick_data[device][3]

    @classmethod
    def joystick_guid(cls, device):
        return cls.joystick_data[device][4]

    @classmethod
    def init(cls):
        if cls.initialized:
            return
        # init can be called more than once (by setting initialized to
        # false, used by refresh function, so we need to clear the lists...
        cls.devices = []
        cls.device_ids = []
        cls.device_names = []
        cls.device_types = []
        cls.device_name_count = {}

        cls.devices.append(Device("none", gettext("No Device"), "none"))
        # cls.devices.append(Device("mouse", _("Mouse"), "mouse"))
        cls.init_fsuae()
        for id, name, type in zip(cls.device_ids, cls.device_names,
                                  cls.device_types):
            cls.devices.append(Device(id, name, type))
        # cls.devices.append(
        #     Device("keyboard", _("Cursor Keys and Right Ctrl/Alt"),
        #            "joystick"))
        cls.initialized = True

    @classmethod
    def refresh(cls):
        cls.initialized = False
        cls.init()
        Signal.broadcast("device_list_updated")

    @classmethod
    def init_fsuae(cls):
        print("finding connected joysticks")
        try:
            p = FSUAEDeviceHelper.start_with_args(
                ["--list"], stdout=subprocess.PIPE)
            joysticks = p.stdout.read()
            p.wait()
        except Exception:
            print("exception while listing joysticks and devices")
            traceback.print_exc()
            return
        print(repr(joysticks))
        # If the character conversion fails, replace will ensure that
        # as much as possible succeeds. The joystick in question will
        # not be pre-selectable in the launcher, but the other ones will
        # work at least.
        joysticks = joysticks.decode("UTF-8", "replace")
        joysticks = [x.strip() for x in joysticks.split("\n") if x.strip()]
        last_joystick = ""
        for line in joysticks:
            if line.startswith("#"):
                continue
            if line.startswith("Buttons:"):
                parts = line.split(" ")
                buttons = int(parts[1])
                hats = int(parts[3])
                axes = int(parts[5])
                balls = int(parts[7])
                guid = parts[9]
                cls.joystick_data[last_joystick] = \
                    buttons, hats, axes, balls, guid

                continue
            device_type, name = line.split(" ", 1)
            # if name.lower() in ["keyboard", "mouse"]:
            #     # these are automatically added
            #     continue
            name_count = cls.device_name_count.get(name, 0) + 1
            cls.device_name_count[name] = name_count
            if name_count > 1:
                name = name + " #" + str(name_count)
            cls.device_ids.append(name)
            name = re.sub("[ ]+", " ", name)
            cls.device_names.append(name)
            if device_type == "J:":
                device_type = "joystick"
                last_joystick = name
            elif device_type == "M:":
                device_type = "mouse"
            elif device_type == "K:":
                # works as an emulated joystick...
                # device_type = "joystick"
                device_type = "keyboard"
            cls.device_types.append(device_type)

        if "--add-dummy-joystick" in sys.argv:
            cls.device_ids.append("Dummy Joystick")
            cls.device_names.append("Dummy Joystick")
            cls.device_types.append("joystick")
            cls.joystick_data["Dummy Joystick"] = \
                1, 0, 2, 0, "c6c1bc29b0124fe6890757bb09ef006f"

    @classmethod
    def get_joystick_names(cls):
        cls.init()
        return cls.device_names[:]

    @classmethod
    def get_device_type(cls, device):
        cls.init()
        for i in range(len(cls.device_names)):
            if cls.device_names[i] == device:
                return cls.device_types[i]
        return ""

    @classmethod
    def is_joystick(cls, device):
        return cls.get_device_type(device) == "joystick"

    @classmethod
    def get_joystick_ids(cls):
        cls.init()
        return cls.device_ids[:]

    @classmethod
    def get_preferred_joysticks(cls):
        prefs = []
        if Settings.get("primary_joystick"):
            prefs.append(create_cmp_id(Settings.get("primary_joystick")))
        if Settings.get("secondary_joystick"):
            prefs.append(create_cmp_id(Settings.get("secondary_joystick")))
        return prefs

    @classmethod
    def get_preferred_gamepads(cls):
        return cls.get_preferred_joysticks()

    @classmethod
    def get_calculated_port_mode(cls, config, port):
        value = config.get("joystick_port_{0}_mode".format(port))
        if not value:
            if port == 0:
                return "mouse"
            elif port == 1:
                if config.get("amiga_model").startswith("CD32"):
                    return "cd32 gamepad"
                else:
                    return "joystick"
            return "nothing"
        return value

    @classmethod
    def get_devices_for_ports(cls, config):
        cls.init()
        ports = [cls.devices[0] for _ in range(5)]
        for device in cls.devices:
            device.port = None
        for p in range(4):
            key = "joystick_port_{0}".format(p)
            value = config.get(key)
            for device in cls.devices:
                if device.id == value:
                    device.port = p
                    break
        # print("-")
        # for device in cls.devices:
        #     print(device.port, device.id)
        # print("-")

        def autofill(port, type):
            mode = config.get("joystick_port_{0}_mode".format(port))
            if not mode:
                mode = cls.get_calculated_port_mode(config, port)
            value = config.get("joystick_port_{0}".format(port))
            if value:
                # specific device chosen
                for device in cls.devices:
                    if device.id == value:
                        ports[port] = device
                        break
                return
            if type == "mouse":
                # print("a", mode)
                if mode != "mouse":
                    return
                # print("b")
                for device in cls.devices:
                    # print("c")
                    if device.type == "mouse" and device.port is None:
                        # print("d")
                        ports[port] = device
                        device.port = port
                        return
            elif type == "joystick":
                if mode == "cd32 gamepad":
                    prefs = cls.get_preferred_joysticks()
                elif mode == "joystick":
                    prefs = cls.get_preferred_gamepads()
                else:
                    return
                # try to find an available preferred device first
                for pref in prefs:
                    for device in cls.devices:
                        if device.cmp_id == pref and device.port is None:
                            ports[port] = device
                            device.port = port
                            return
                # find first suitable device
                for device in cls.devices:
                    if device.type == "joystick" and device.port is None:
                        ports[port] = device
                        device.port = port
                        return
                for device in cls.devices:
                    if device.type == "keyboard" and device.port is None:
                        ports[port] = device
                        device.port = port
                        return

        for p in [0, 1, 2, 3]:
            autofill(p, "mouse")
        for p in [1, 0, 3, 2]:
            autofill(p, "joystick")
        return ports

    @classmethod
    def get_device_for_port(cls, config, port):
        return cls.get_devices_for_ports(config)[port]



import os
from pkg_resources import resource_filename
from .Application import Application


class Resources(object):

    def __init__(self, package, subdir=""):
        self.package = package
        self.subdir = subdir
        self.resource_sha1s = {}

    def resource_name(self, resource):
        if self.subdir:
            # return os.path.join(self.subdir, resource)
            return self.subdir + "/" + resource
        return resource

    def stream(self, resource):
        # resource_name = self.resource_name(resource)
        # return resource_stream(self.package, str_path(resource_name))
        # try:
        #     return resource_stream(self.package, Paths.encode(resource_name))
        # except Exception:
        #     return open(self.path(resource), "rb")
        try:
            return open(self.path(resource), "rb")
        except FileNotFoundError:
            raise LookupError(resource)

    def path(self, resource):
        print("looking up resource", resource)
        print("- package:", self.package)
        application = Application.get_instance()
        resource_name = self.resource_name(resource)
        print("resource_name", resource_name)

        relative_path = os.path.join(self.package, resource_name)
        try:
            return application.data_file(relative_path)
        except LookupError:
            pass

        try:
            print("resource_filename(\"{0}\", \"{1}\")".format(
                self.package, resource_name))
            return resource_filename(self.package, resource_name)
            # return resource_filename(
            #     self.package, Paths.encode(resource_name))
        except Exception:
            pass

        raise LookupError(
            "Cannot find resource {0}".format(repr(resource)))

        # try:
        #     sha1 = self.resource_sha1s[resource]
        # except KeyError:
        #     stream = self.stream(resource)
        #     data = stream.read()
        #     sha1 = hashlib.sha1(data).hexdigest()
        #     self.resource_sha1s[resource] = sha1
        # else:
        #     data = None

        # _, ext = os.path.splitext(resource)
        # ext = ext or ".bin"
        # cache_path = os.path.join(application.cache_dir(), "Temp", sha1 + ext)
        # if not os.path.exists(cache_path):
        #     if not os.path.exists(os.path.dirname(cache_path)):
        #         os.makedirs(os.path.dirname(cache_path))
        #     if data is None:
        #         stream = self.stream(resource)
        #         data = stream.read()
        #     with open(cache_path + ".partial", "wb") as f:
        #         f.write(data)
        #     os.rename(cache_path + ".partial", cache_path)
        # return cache_path

        # if not path or not os.path.exists(path):
        #     raise LookupError(
        #         "Cannot find resource {0}".format(repr(resource)))
        # return path


# import os
# import sys
# import logging
# from fengestad.fs import encode_path, memoize
#
# logger = logging.getLogger('fengestad.game_center')
#
# def _(msg):
#     return six.text_type(msg)
#
# def ngettext(n, msg1, msg2):
#     return six.text_type(msg1) if n == 1 else six.text_type(msg2)
#
# class Resources(object):
#     def __init__(self, package_or_requirement):
#         self.req = package_or_requirement
#
#     def resource_pil_image(self, resource_name):
#         resource_name = encode_path(u'res/' + resource_name)
#         return resource_pil_image(self.req, resource_name)
#
#     def resource_stream(self, resource_name):
#         resource_name = encode_path(u'res/' + resource_name)
#         return resource_stream(self.req, resource_name)
#
#     def resource_filename(self, resource_name):
#         print >> sys.stderr, resource_name
#         try:
#             resource_name_2 = encode_path(u'res/' + resource_name)
#             path = resource_filename(self.req, resource_name_2)
#             print >> sys.stderr, path
#             if not os.path.exists(path):
#                 raise Exception("Cannot find resource {0}".format(
#                         repr(resource_name)))
#             return path
#         except Exception:
#             # for Windows
#             path = os.path.join("share", "fs-game-center", resource_name)
#             print >> sys.stderr, os.getcwd()
#             print >> sys.stderr, path
#             if os.path.exists(path):
#                 return path
#             raise Exception(
#                 "Cannot find resource {0}".format(repr(resource_name)))

# def resource_pil_image(package_or_requirement, resource_name):
#     #print("resource_pil_image", package_or_requirement, resource_name)
#     stream = resource_stream(package_or_requirement, resource_name)
#     #from PIL import Image
#     import Image
#     return Image.open(stream)
#
#
# def resource_icon_stream(package_or_requirement, name, size):
#     #print(name, size)
#     resource_name = encode_path(u'res/icons/%dx%d/%s.png' % (
#         size, size, name))
#     return resource_stream(package_or_requirement, resource_name)
#
#
# def resource_icon_pil(package_or_requirement, name, size):
#     #print("resource_icon_pil", name, size)
#     #path = os.path.join(self.path, "icons",
#     #        "%dx%d" % (size, size), name + ".png")
#     #if os.path.isfile(path):
#     #    return Image.open(path)
#     stream = resource_icon_stream(package_or_requirement, name, size)
#     #from PIL import Image
#     import Image
#     return Image.open(stream)
#
#
# resources = Resources("fengestad.game_center")



import fsui.qt
from .Widget import Widget


class ImageView(fsui.qt.QLabel, Widget):

    def __init__(self, parent, image):
        fsui.qt.QLabel.__init__(self, parent.get_container())
        self.init_widget(parent)
        self.setPixmap(image.qpixmap)

    def set_image(self, image):
        self.setPixmap(image.qpixmap)



import json
from fsgs.platform import PlatformHandler
from fsgs.hatari.atari_st import AtariSTRunner
from .loader import SimpleLoader


class AtariSTPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Atari ST"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return AtariSTLoader(fsgs)

    def get_runner(self, fsgs):
        return AtariSTRunner(fsgs)


class AtariSTLoader(SimpleLoader):

    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        # assert len(file_list) == 1
        if file_list[0]["name"].endswith(".st"):
            self.config["floppy_drive_0"] = "sha1://{0}/{1}".format(
                file_list[0]["sha1"], file_list[0]["name"])



'''OpenGL extension VERSION.GL_4_3

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_4_3 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_4_3.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_4_3 import *
from OpenGL.raw.GL.VERSION.GL_4_3 import _EXTENSION_NAME

def glInitGl43VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION

from OpenGL.GL.ARB.arrays_of_arrays import *
from OpenGL.GL.ARB.fragment_layer_viewport import *
from OpenGL.GL.ARB.shader_image_size import *
from OpenGL.GL.ARB.ES3_compatibility import *
from OpenGL.GL.ARB.clear_buffer_object import *
from OpenGL.GL.ARB.compute_shader import *
from OpenGL.GL.ARB.copy_image import *
# Extension registry no longer defines these extensions?
#from OpenGL.GL.ARB.debug_group import *
#from OpenGL.GL.ARB.debug_label import *
#from OpenGL.GL.ARB.debug_output2 import *
from OpenGL.GL.KHR.debug import *
from OpenGL.GL.ARB.explicit_uniform_location import *
from OpenGL.GL.ARB.framebuffer_no_attachments import *
from OpenGL.GL.ARB.internalformat_query2 import *
from OpenGL.GL.ARB.invalidate_subdata import *
from OpenGL.GL.ARB.multi_draw_indirect import *
from OpenGL.GL.ARB.program_interface_query import *
from OpenGL.GL.ARB.robust_buffer_access_behavior import *
from OpenGL.GL.ARB.shader_storage_buffer_object import *
from OpenGL.GL.ARB.stencil_texturing import *
from OpenGL.GL.ARB.texture_buffer_range import *
from OpenGL.GL.ARB.texture_query_levels import *
from OpenGL.GL.ARB.texture_storage_multisample import *
from OpenGL.GL.ARB.texture_view import *
from OpenGL.GL.ARB.vertex_attrib_binding import *



"""Implementation of the special "glGet" functions

For comparison, here's what a straightforward implementation looks like:

    def glGetDoublev( pname ):
        "Natural writing of glGetDoublev using standard ctypes"
        output = c_double*sizes.get( pname )
        result = output()
        result = platform.PLATFORM.GL.glGetDoublev( pname, byref(result) )
        return Numeric.array( result )
"""
from OpenGL import wrapper
from OpenGL.raw.GL.VERSION import GL_1_1 as _simple
from OpenGL.raw.GL import _glgets
import ctypes
GLenum = ctypes.c_uint
GLsize = GLsizei = ctypes.c_int

__all__ = (
    'glGetBoolean','glGetBooleanv','glGetInteger','glGetIntegerv',
    'glGetFloat','glGetFloatv','glGetDouble','glGetDoublev',
    'glGetString',
    'glGetLightfv','glGetLightiv',
    'glGetMaterialfv','glGetMaterialiv',
    'glGetPixelMapfv','glGetPixelMapusv','glGetPixelMapuiv',
    'glGetPolygonStipple', 'glGetPolygonStippleub',
    'glGetTexEnviv','glGetTexEnvfv',
    'glGetTexGenfv','glGetTexGeniv','glGetTexGendv',
    'glGetTexLevelParameteriv',
    'glGetTexLevelParameterfv',
    'glGetTexParameterfv',
    'glGetTexParameteriv',
)

glGetString = _simple.glGetString
glGetString.restype = ctypes.c_char_p
glGetString.__doc__ = """glGetString( constant ) -> Current string value"""

GL_GET_SIZES = _glgets._glget_size_mapping
def addGLGetConstant( constant, arraySize ):
    """Add a glGet* constant to return an output array of correct size"""
    GL_GET_SIZES[ constant ] = arraySize

glGetDouble = glGetDoublev = wrapper.wrapper(_simple.glGetDoublev).setOutput(
    "data",GL_GET_SIZES, "pname", orPassIn=True
)
glGetFloat = glGetFloatv = wrapper.wrapper(_simple.glGetFloatv).setOutput(
    "data",GL_GET_SIZES, "pname", orPassIn=True
)
glGetBoolean = glGetBooleanv = glGetInteger = glGetIntegerv = wrapper.wrapper(_simple.glGetIntegerv).setOutput(
    "data",GL_GET_SIZES, "pname", orPassIn=True
)

GL_GET_LIGHT_SIZES = {
    # glGetLightXv
    _simple.GL_AMBIENT                               : (4,),
    _simple.GL_DIFFUSE                               : (4,),
    _simple.GL_SPECULAR                              : (4,),
    _simple.GL_POSITION                              : (4,),
    _simple.GL_SPOT_DIRECTION                        : (3,),
    _simple.GL_SPOT_EXPONENT                         : (1,),
    _simple.GL_SPOT_CUTOFF                           : (1,),
    _simple.GL_CONSTANT_ATTENUATION                  : (1,),
    _simple.GL_LINEAR_ATTENUATION                    : (1,),
    _simple.GL_QUADRATIC_ATTENUATION                 : (1,),
} # end of sizes
glGetLightfv = wrapper.wrapper(_simple.glGetLightfv).setOutput(
    "params",GL_GET_LIGHT_SIZES, "pname", orPassIn=True
)
glGetLightiv = wrapper.wrapper(_simple.glGetLightiv).setOutput(
    "params",GL_GET_LIGHT_SIZES, "pname", orPassIn=True
)

GL_GET_MATERIAL_SIZES = {
    _simple.GL_AMBIENT: (4,),
    _simple.GL_DIFFUSE: (4,),
    _simple.GL_SPECULAR: (4,),
    _simple.GL_EMISSION: (4,),
    _simple.GL_SHININESS: (1,),
    _simple.GL_COLOR_INDEXES: (3,)
}
glGetMaterialfv = wrapper.wrapper(_simple.glGetMaterialfv).setOutput(
    "params",GL_GET_MATERIAL_SIZES, "pname", orPassIn=True
)
glGetMaterialiv = wrapper.wrapper(_simple.glGetMaterialiv).setOutput(
    "params",GL_GET_MATERIAL_SIZES, "pname", orPassIn=True
)
PIXEL_MAP_SIZE_CONSTANT_MAP = {
    _simple.GL_PIXEL_MAP_A_TO_A: _simple.GL_PIXEL_MAP_A_TO_A_SIZE,
    _simple.GL_PIXEL_MAP_B_TO_B: _simple.GL_PIXEL_MAP_B_TO_B_SIZE,
    _simple.GL_PIXEL_MAP_G_TO_G: _simple.GL_PIXEL_MAP_G_TO_G_SIZE,
    _simple.GL_PIXEL_MAP_I_TO_A: _simple.GL_PIXEL_MAP_I_TO_A_SIZE,
    _simple.GL_PIXEL_MAP_I_TO_B: _simple.GL_PIXEL_MAP_I_TO_B_SIZE,
    _simple.GL_PIXEL_MAP_I_TO_G: _simple.GL_PIXEL_MAP_I_TO_G_SIZE,
    _simple.GL_PIXEL_MAP_I_TO_I: _simple.GL_PIXEL_MAP_I_TO_I_SIZE,
    _simple.GL_PIXEL_MAP_I_TO_R: _simple.GL_PIXEL_MAP_I_TO_R_SIZE,
    _simple.GL_PIXEL_MAP_R_TO_R: _simple.GL_PIXEL_MAP_R_TO_R_SIZE,
    _simple.GL_PIXEL_MAP_S_TO_S: _simple.GL_PIXEL_MAP_S_TO_S_SIZE,
}
def GL_GET_PIXEL_MAP_SIZE( pname ):
    """Given a pname, lookup the size using a glGet query..."""
    constant = PIXEL_MAP_SIZE_CONSTANT_MAP[ pname ]
    return glGetIntegerv( constant )
glGetPixelMapfv = wrapper.wrapper(_simple.glGetPixelMapfv).setOutput(
    "values",GL_GET_PIXEL_MAP_SIZE, "map", orPassIn=True
)
glGetPixelMapuiv = wrapper.wrapper(_simple.glGetPixelMapuiv).setOutput(
    "values",GL_GET_PIXEL_MAP_SIZE, "map", orPassIn=True
)
glGetPixelMapusv = wrapper.wrapper(_simple.glGetPixelMapusv).setOutput(
    "values",GL_GET_PIXEL_MAP_SIZE, "map", orPassIn=True
)

# 32 * 32 bits
POLYGON_STIPPLE_SIZE = (32*32//8,)
glGetPolygonStipple = glGetPolygonStippleub = wrapper.wrapper(_simple.glGetPolygonStipple).setOutput(
    "mask",POLYGON_STIPPLE_SIZE, orPassIn=True
)
GL_GET_TEX_ENV_SIZES = {
    _simple.GL_TEXTURE_ENV_MODE: (1,),
    _simple.GL_TEXTURE_ENV_COLOR: (4,),
}
glGetTexEnvfv = wrapper.wrapper(_simple.glGetTexEnvfv).setOutput(
    "params",GL_GET_TEX_ENV_SIZES, 'pname', orPassIn=True
)
glGetTexEnviv = wrapper.wrapper(_simple.glGetTexEnviv).setOutput(
    "params",GL_GET_TEX_ENV_SIZES, 'pname', orPassIn=True
)
GL_GET_TEX_GEN_SIZES = {
    _simple.GL_TEXTURE_GEN_MODE: (1,),
    _simple.GL_OBJECT_PLANE: (4,),
    _simple.GL_EYE_PLANE: (4,),
}
glGetTexGendv = wrapper.wrapper(_simple.glGetTexGendv).setOutput(
    "params",GL_GET_TEX_GEN_SIZES, 'pname', orPassIn=True
)
glGetTexGenfv = wrapper.wrapper(_simple.glGetTexGenfv).setOutput(
    "params",GL_GET_TEX_GEN_SIZES, 'pname', orPassIn=True
)
glGetTexGeniv = wrapper.wrapper(_simple.glGetTexGeniv).setOutput(
    "params",GL_GET_TEX_GEN_SIZES, 'pname', orPassIn=True
)

glGetTexLevelParameterfv = wrapper.wrapper(_simple.glGetTexLevelParameterfv).setOutput(
    "params",(1,), orPassIn=True
)
glGetTexLevelParameteriv = wrapper.wrapper(_simple.glGetTexLevelParameteriv).setOutput(
    "params",(1,), orPassIn=True
)
TEX_PARAMETER_SIZES = {
    _simple.GL_TEXTURE_MAG_FILTER: (1,),
    _simple.GL_TEXTURE_MIN_FILTER: (1,),
    _simple.GL_TEXTURE_WRAP_S: (1,),
    _simple.GL_TEXTURE_WRAP_T: (1,),
    _simple.GL_TEXTURE_BORDER_COLOR: (4,),
    _simple.GL_TEXTURE_PRIORITY: (1,),
    _simple.GL_TEXTURE_RESIDENT: (1,)
}
def addGLGetTexParameterConstant( constant, arraySize ):
    TEX_PARAMETER_SIZES[constant] = arraySize

glGetTexParameterfv = wrapper.wrapper(_simple.glGetTexParameterfv).setOutput(
    "params",TEX_PARAMETER_SIZES, 'pname', orPassIn=True
)
glGetTexParameteriv = wrapper.wrapper(_simple.glGetTexParameteriv).setOutput(
    "params",TEX_PARAMETER_SIZES, 'pname', orPassIn=True
)



import os
import sys
import subprocess

print("a")
if os.path.exists("c:/signtool.py"):
    assert os.system("python c:/signtool.py \"{}\"".format(
        sys.argv[1])) == 0
    # p = subprocess.Popen(["python", "c:/signtool.py", sys.argv[1]])
    # assert p.wait() == 0



from fsbc.signal import Signal
from .ContextAware import ContextAware


class SignalContext(ContextAware):

    def __init__(self, context):
        ContextAware.__init__(self, context)

    def connect(self, signal, listener):
        signal = "fsgs:" + signal
        Signal(signal).connect(listener)

    def disconnect(self, signal, listener):
        signal = "fsgs:" + signal
        Signal(signal).disconnect(listener)

    def notify(self, signal, args):
        signal = "fsgs:" + signal
        Signal(signal).notify(*args)

    def process(self):
        Signal.process_all_signals()



import fsui as fsui
from ..Config import Config
from ..I18N import gettext
from ..Settings import Settings
from fs_uae_launcher.ui.newbutton import NewButton
from fs_uae_launcher.ui.savebutton import SaveButton


class ConfigGroup(fsui.Group):

    def __init__(self, parent, new_button=True):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        # heading_label = fsui.HeadingLabel(self, _("Configuration"))
        # self.layout.add(heading_label, margin=10)
        # self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        # dummy label for sizing purposes
        # label = fsui.HeadingLabel(self, "")
        # hori_layout.add(label, margin_top=10, margin_bottom=10)

        if new_button:
            label_stand_in = fsui.Panel(self)
            tw, th = label_stand_in.measure_text(gettext("Configuration"))
            label_stand_in.set_min_height(th)
            hori_layout.add(label_stand_in, margin_top=10, margin_bottom=10)

            hori_layout.add(NewButton(self), margin_left=10, margin_right=10)

        # self.open_button = IconButton(self, "open_button.png")
        # self.open_button.set_tooltip(_("Open Configuration"))
        # self.open_button.disable()
        # self.open_button.activated.connect(self.on_open_button)
        # hori_layout.add(self.open_button, margin=10)

        self.config_name_field = fsui.TextField(self)
        hori_layout.add(self.config_name_field, expand=True, margin=10,
                        margin_top=0, margin_bottom=0)

        hori_layout.add(SaveButton(self), margin_left=10, margin_right=10)

        # self.save_button = IconButton(self, "save_button.png")
        # # self.save_button.disable()
        # self.save_button.set_tooltip(gettext("Save Configuration"))
        # self.save_button.activated.connect(self.on_save_button)
        # hori_layout.add(self.save_button, margin=10,
        #                 margin_top=0, margin_bottom=0)

        self.on_setting("config_name", Settings.get("config_name"))
        self.config_name_field.on_change = self.on_config_name_change

        # Config.add_listener(self)
        Settings.add_listener(self)

        # self.on_setting("config_changed", Settings.get("config_changed"))
        # self.on_config("__changed", Config.get("__changed"))

    def on_destroy(self):
        # Config.remove_listener(self)
        Settings.remove_listener(self)

    # def on_config(self, key, value):
    #     if key == "__changed":
    #         self.save_button.enable(value == "1")

    def on_setting(self, key, value):
        # if key == "config_changed":
        #     self.save_button.enable(value == "1")
        if key == "config_name":
            if value != self.config_name_field.get_text().strip():
                self.config_name_field.set_text(value)

    def on_config_name_change(self):
        text = self.config_name_field.get_text().strip()
        Settings.set("config_name", text)
        # Settings.set("config_changed", "1")
        Config.set("__changed", "1")
        # FIXME: remove
        # Config.set("title", text)

    @staticmethod
    def new_config():
        NewButton.new_config()



'''OpenGL extension SUN.get_transparent_index

This module customises the behaviour of the 
OpenGL.raw.GLX.SUN.get_transparent_index to provide a more 
Python-friendly API

Overview (from the spec)
	
	Gets the transparent pixel index for an overlay/underlay window pair.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/SUN/get_transparent_index.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper

import ctypes
from OpenGL.raw.GLX import _types
from OpenGL.raw.GLX.SUN.get_transparent_index import *
from OpenGL.raw.GLX.SUN.get_transparent_index import _EXTENSION_NAME

def glInitGetTransparentIndexSUN():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


'''OpenGL extension VERSION.GLES2_2_0

This module customises the behaviour of the 
OpenGL.raw.GLES2.VERSION.GLES2_2_0 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GLES2_2_0.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper

import ctypes
from OpenGL.raw.GLES2 import _types
from OpenGL.raw.GLES2.VERSION.GLES2_2_0 import *
from OpenGL.raw.GLES2.VERSION.GLES2_2_0 import _EXTENSION_NAME

def glInitGles220VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
from OpenGL import converters
from OpenGL.lazywrapper import lazy as _lazy

glShaderSource = platform.createExtensionFunction(
    'glShaderSource', dll=platform.PLATFORM.GLES2,
    resultType=None,
    argTypes=(_types.GLhandle, _types.GLsizei, ctypes.POINTER(ctypes.c_char_p), arrays.GLintArray,),
    doc = 'glShaderSource( GLhandle(shaderObj),[bytes(string),...]) -> None',
    argNames = ('shaderObj', 'count', 'string', 'length',),
    extension = _EXTENSION_NAME,
)

conv = converters.StringLengths( name='string' )
glShaderSource = wrapper.wrapper(
    glShaderSource
).setPyConverter(
    'count' # number of strings
).setPyConverter(
    'length' # lengths of strings
).setPyConverter(
    'string', conv.stringArray
).setCResolver(
    'string', conv.stringArrayForC,
).setCConverter(
    'length', conv,
).setCConverter(
    'count', conv.totalCount,
)
try:
    del conv
except NameError as err:
    pass

glGetShaderiv = wrapper.wrapper( glGetShaderiv ).setOutput(
    'params',
    size=(1,), 
    arrayType=arrays.GLintArray,
    orPassIn = True,
)


@_lazy( glGetShaderInfoLog )
def glGetShaderInfoLog( baseOperation, obj ):
    """Retrieve the shader's error messages as a Python string

    returns string which is '' if no message
    """
    length = int(glGetShaderiv(obj, GL_INFO_LOG_LENGTH))
    if length > 0:
        log = ctypes.create_string_buffer(length)
        baseOperation(obj, length, None, log)
        return log.value.strip(_NULL_8_BYTE) # null-termination
    return ''
@_lazy( glGetProgramInfoLog )
def glGetProgramInfoLog( baseOperation, obj ):
    """Retrieve the shader program's error messages as a Python string

    returns string which is '' if no message
    """
    length = int(glGetProgramiv(obj, GL_INFO_LOG_LENGTH))
    if length > 0:
        log = ctypes.create_string_buffer(length)
        baseOperation(obj, length, None, log)
        return log.value.strip(_NULL_8_BYTE) # null-termination
    return ''

@_lazy( glGetAttachedShaders )
def glGetAttachedShaders( baseOperation, obj ):
    """Retrieve the attached objects as an array of GLhandle instances"""
    length= glGetProgramiv( obj, GL_ATTACHED_SHADERS )
    if length > 0:
        storage = arrays.GLuintArray.zeros( (length,))
        baseOperation( obj, length, None, storage )
        return storage
    return arrays.GLuintArray.zeros( (0,))


@_lazy( glGetShaderSource )
def glGetShaderSource( baseOperation, obj ):
    """Retrieve the program/shader's source code as a Python string

    returns string which is '' if no source code
    """
    length = int(glGetShaderiv(obj, GL_OBJECT_SHADER_SOURCE_LENGTH))
    if length > 0:
        source = ctypes.create_string_buffer(length)
        baseOperation(obj, length, None, source)
        return source.value.strip(_NULL_8_BYTE) # null-termination
    return ''

@_lazy( glGetActiveUniform )
def glGetActiveUniform(baseOperation,program, index):
    """Retrieve the name, size and type of the uniform of the index in the program"""
    max_index = int(glGetProgramiv( program, GL_OBJECT_ACTIVE_UNIFORMS ))
    length = int(glGetProgramiv( program, GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH))
    if index < max_index and index >= 0:
        if length > 0:
            name = ctypes.create_string_buffer(length)
            size = arrays.GLintArray.zeros( (1,))
            gl_type = arrays.GLenumArray.zeros( (1,))
            namelen = arrays.GLsizeiArray.zeros( (1,))
            baseOperation(program, index, length, namelen, size, gl_type, name)
            return name.value[:int(namelen[0])], size[0], gl_type[0]
        raise ValueError( """No currently specified uniform names""" )
    raise IndexError( 'Index %s out of range 0 to %i' % (index, max_index - 1, ) )

@_lazy( glGetUniformLocation )
def glGetUniformLocation( baseOperation, program, name ):
    """Check that name is a string with a null byte at the end of it"""
    if not name:
        raise ValueError( """Non-null name required""" )
    name = as_8_bit( name )
    if name[-1] != _NULL_8_BYTE:
        name = name + _NULL_8_BYTE
    return baseOperation( program, name )
@_lazy( glGetAttribLocation )
def glGetAttribLocation( baseOperation, program, name ):
    """Check that name is a string with a null byte at the end of it"""
    if not name:
        raise ValueError( """Non-null name required""" )
    
    name = as_8_bit( name )
    if name[-1] != _NULL_8_BYTE:
        name = name + _NULL_8_BYTE
    return baseOperation( program, name )

@_lazy( glVertexAttribPointer )
def glVertexAttribPointer(
    baseOperation, index, size, type,
    normalized, stride, pointer,
):
    """Set an attribute pointer for a given shader (index)

    index -- the index of the generic vertex to bind, see
        glGetAttribLocation for retrieval of the value,
        note that index is a global variable, not per-shader
    size -- number of basic elements per record, 1,2,3, or 4
    type -- enum constant for data-type
    normalized -- whether to perform int to float
        normalization on integer-type values
    stride -- stride in machine units (bytes) between
        consecutive records, normally used to create
        "interleaved" arrays
    pointer -- data-pointer which provides the data-values,
        normally a vertex-buffer-object or offset into the
        same.

    This implementation stores a copy of the data-pointer
    in the contextdata structure in order to prevent null-
    reference errors in the renderer.
    """
    array = ArrayDatatype.asArray( pointer, type )
    key = ('vertex-attrib',index)
    contextdata.setValue( key, array )
    return baseOperation(
        index, size, type,
        normalized, stride,
        ArrayDatatype.voidDataPointer( array )
    )




from fs_uae_launcher.ui.behaviors.amigaenablebehavior import AmigaEnableBehavior
import fsui as fsui
from ..CDManager import CDManager
from ..FloppyManager import FloppyManager
from ..I18N import gettext
from .FloppySelector import FloppySelector


class FloppiesGroup(fsui.Group):

    def __init__(self, parent, drives=2, cd_mode=False):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.cd_mode = cd_mode
        self.num_drives = drives

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        if cd_mode:
            self.label = fsui.HeadingLabel(self, gettext("CD-ROM Drive"))
        else:
            self.label = fsui.HeadingLabel(self, gettext("Floppy Drives"))
        hori_layout.add(self.label, margin=10)
        hori_layout.add_spacer(0, expand=True)

        self.multi_select_button = fsui.Button(
            self, gettext("Multi-Select..."))
        if self.cd_mode:
            self.multi_select_button.set_tooltip(
                gettext("Add Multiple CD-ROMs at Once"))
        else:
            self.multi_select_button.set_tooltip(
                gettext("Add Multiple Floppies at Once"))
        AmigaEnableBehavior(self.multi_select_button)
        self.multi_select_button.activated.connect(self.on_multi_select_button)

        hori_layout.add(self.multi_select_button, margin_right=10)

        self.layout.add_spacer(0)

        self.selectors = []
        for i in range(drives):
            selector = FloppySelector(parent, i)
            if cd_mode:
                selector.set_cd_mode(True)
            self.selectors.append(selector)
            self.layout.add(selector, fill=True, margin=10)

    def on_multi_select_button(self):
        if self.cd_mode:
            CDManager.multiselect(self.get_window())
        else:
            FloppyManager.multiselect(self.get_window())

    def update_heading_label(self):
        if self.cd_mode:
            if self.num_drives > 1:
                self.label.set_text(gettext("CD-ROM Drives"))
            else:
                self.label.set_text(gettext("CD-ROM Drive"))
        else:
            self.label.set_text(gettext("Floppy Drives"))



import fsui


volumes = None


def get_vfs_volumes():
    global volumes
    if volumes is None:
        volumes = [BuiltinItem(
            "Workspace", "Workspace:", default_items["Workspace"])]
    return volumes[:]


def get_vfs_item(uri):
    volume_name, path = uri.split(":")
    for volume in get_vfs_volumes():
        if not volume.name() == volume_name:
            continue
        item = volume
        break
    else:
        return None
    parts = path.split("/")
    for part in parts:
        item = item.item(part)
    return item


class VFSIcon(object):
    pass


class VFSKnownIcon(VFSIcon):

    def __init__(self, name):
        self.name = name

    def image(self):
        return fsui.Image("fs_uae_workspace:res/48/" + self.name + ".png")


class VFSNoIcon(VFSKnownIcon):

    __instance = None

    @classmethod
    def instance(cls):
        if cls.__instance is None:
            cls.__instance = VFSNoIcon("no_icon")
        return cls.__instance


class VFSItem(object):

    def __init__(self, uri):
        self._uri = uri

    def icon(self):
        return VFSNoIcon.instance()

    def children(self):
        return []

    def uri(self):
        return self._uri

    def item(self, name):
        raise KeyError(name)

    def name(self):
        raise NotImplementedError()

    def display_name(self):
        return self.name()

    # def open(self, mode):
    #     raise NotImplementedError()


class VFSVolume(VFSItem):
    pass


class BuiltinItem(VFSItem):

    def __init__(self, name, uri, node):
        VFSItem.__init__(self, uri)
        self._name = name
        self._uri = uri
        self._node = node

    def name(self):
        return self._name

    def open(self, args=[]):
        if "app" in self._node:
            _temp = __import__(self._node["app"], globals(), locals(),
                               ["application"])
            application = _temp.application(self.uri(), args)

    def content_type(self):
        if "app" in self._node:
            return "application/x-fs-uae-application"
        else:
            return "application/x-fs-uae-vfolder"

    def uri(self):
        return self._uri

    def icon(self):
        return VFSKnownIcon(self._node["icon"])

    def children(self):
        return [x for x in self._node["children"]]

    def item(self, name):
        uri = self.uri()
        if uri.endswith(":"):
            uri = self._uri + name
        else:
            uri = self._uri + "/" + name
        return BuiltinItem(name, uri, self._node["children"][name])


class VFSDesktopItem(VFSItem):

    def __init__(self):
        # noinspection PyDictCreation
        self._items = {}
        # self._items["FS-UAE (Old)"] = FSUAEVolume()

        self._items["Workspace"] = BuiltinItem(
            "Workspace", "Workspace:", default_items["Workspace"])
        self._items["Files"] = BuiltinItem(
            "Files", "Files:", default_items["Files"])
        self._items["Ubuntu"] = BuiltinItem(
            "Ubuntu", "Ubuntu:", default_items["Ubuntu"])

        self._items["Clock"] = BuiltinItem(
            "Clock", "Workspace:Utilities/Clock",
            default_items["Workspace"]["children"]["Utilities"][
                "children"]["Clock"])
        self._items["Language"] = BuiltinItem(
            "Language", "Workspace:Utilities/Language",
            default_items["Workspace"]["children"]["Prefs"][
                "children"]["Language"])

        self._items["ADFCreator"] = BuiltinItem(
            "Language", "Workspace:Tools/ADFCreator",
            default_items["Workspace"]["children"]["Tools"][
                "children"]["ADFCreator"])
        self._items["HDFCreator"] = BuiltinItem(
            "Language", "Workspace:Tools/HDFCreator",
            default_items["Workspace"]["children"]["Tools"][
                "children"]["HDFCreator"])

        # self._items["Floppies"] = FloppiesVolume()
        # self._items["Configs"] = ConfigsVolume()

    def name(self):
        return "FS-UAE Workbench"

    # def icon(self):
    #     return VFSKnownIcon("fs-uae")

    def children(self):
        return self._items.keys()

    def item(self, name):
        return self._items[name]


class FileSystemDirectory(VFSItem):

    def __init__(self, uri, path):
        VFSItem.__init__(self, uri)


class FileSystemVolume(FileSystemDirectory):

    def __init__(self, uri, name, path):
        VFSItem.__init__(self, uri)
        pass


class FSUAEPreferences(VFSItem):

    def icon(self):
        return VFSKnownIcon("preferences")


class FSUAETools(VFSItem):

    def icon(self):
        return VFSKnownIcon("tools")


class FSUAELauncher(VFSItem):

    def icon(self):
        return VFSKnownIcon("fs-uae-launcher")


class FSUAEPlugins(VFSItem):

    def icon(self):
        return VFSKnownIcon("folder-plugins")


class FSUAEGameCenter(VFSItem):

    def icon(self):
        return VFSKnownIcon("arcade")


class FSUAENetplay(VFSItem):

    def icon(self):
        return VFSKnownIcon("netplay")


class FSUAEAmiga(VFSItem):

    def icon(self):
        return VFSKnownIcon("amiga-computer")


class FSUAEKickstarts(VFSItem):

    def icon(self):
        return VFSKnownIcon("kickstarts")


class FSUAEConfigs(VFSItem):

    def icon(self):
        return VFSKnownIcon("folder")


class FSUAEVolume(VFSItem):

    _item_map = {
        "Preferences": FSUAEPreferences,
        "Tools": FSUAETools,
        "Launcher": FSUAELauncher,
        "Plugins": FSUAEPlugins,
        "Arcade": FSUAEGameCenter,
        "Netplay": FSUAENetplay,
        "Models": FSUAEAmiga,
        "Kickstarts": FSUAEKickstarts,
        "Configs": FSUAEConfigs,
    }

    def __init__(self):
        VFSItem.__init__(self, "FS-UAEOld:")

    def name(self):
        return "FS-UAE"

    def icon(self):
        return VFSKnownIcon("fs-uae-volume")

    def children(self):
        return self._item_map.keys()

    def item(self, name):
        item_uri = self.uri() + name
        return self._item_map[name](item_uri)


class FloppiesVolume(VFSItem):

    def __init__(self):
        VFSItem.__init__(self, "Floppies:")

    def name(self):
        return "Floppies"

    def icon(self):
        return VFSKnownIcon("floppies-volume")

    def children(self):
        return []


class ConfigsVolume(VFSItem):

    def __init__(self):
        VFSItem.__init__(self, "Configs:")

    def name(self):
        return "Configs"

    def icon(self):
        return VFSKnownIcon("configs-volume")

    def children(self):
        return []


default_items = {
    "Workspace": {
        "icon": "fs-uae-volume",
        "children": {
            "Arcade": {
                "icon": "arcade",
                "app": "fs_uae_workspace.apps.arcade_app",
            },
            "Launcher": {
                "icon": "fs-uae-launcher",
                "app": "fs_uae_workspace.apps.launcher_app",
            },
            "Netplay": {
                "icon": "netplay",
                "app": "fs_uae_workspace.apps.netplay_app",
            },
            "Plugins": {
                "icon": "plugins-folder",
            },
            "Prefs": {
                "icon": "preferences",
                "children": {
                    "Audio": {
                        "icon": "audio-settings",
                        "app": "fs_uae_workspace.prefs.audio",
                    },
                    "Experimental": {
                        "icon": "settings",
                        "app": "fs_uae_workspace.prefs.experimental_features",
                    },
                    "Filters": {
                        "icon": "filter-settings",
                        "app": "fs_uae_workspace.prefs.filter",
                    },
                    "GameDB": {
                        "icon": "database-settings",
                        "app": "fs_uae_workspace.prefs.game_database",
                    },
                    "Keyboard": {
                        "icon": "keyboard-settings",
                        "app": "fs_uae_workspace.prefs.keyboard",
                    },
                    "Joysticks": {
                        "icon": "joystick-settings",
                        "app": "fs_uae_workspace.prefs.joystick",
                    },
                    "Language": {
                        "icon": "language-settings",
                        "app": "fs_uae_workspace.prefs.language",
                    },
                    "Mice": {
                        "icon": "mouse-settings",
                        "app": "fs_uae_workspace.prefs.mouse",
                    },
                    "Netplay": {
                        "icon": "netplay-settings",
                        "app": "fs_uae_workspace.prefs.netplay",
                    },
                    # "OpenGL": {
                    #     "icon": "setting",
                    #     "app": "fs_uae_workspace.prefs.opengl",
                    # },
                    "Scan": {
                        "icon": "indexing-settings",
                        "app": "fs_uae_workspace.prefs.scan",
                    },
                    "User": {
                        "icon": "preferences",
                        "children": {
                            "Login": {
                                "icon": "preferences",
                                "app": "fs_uae_workspace.apps.login",
                            },
                            "Logout": {
                                "icon": "preferences",
                                "app": "fs_uae_workspace.apps.logout",
                            },
                        },
                    },
                    "Video": {
                        "icon": "video-settings",
                        "app": "fs_uae_workspace.prefs.video",
                    },
                },
            },
            "Tools": {
                "icon": "tools",
                "children": {
                    "ADFCreator": {
                        "icon": "floppy",
                        "app": "fs_uae_workspace.apps.adf_creator_app",
                    },
                    "Calculator": {
                        "icon": "calculator",
                        "app": "fs_uae_workspace.apps.calculator_app",
                    },
                    "HDFCreator": {
                        "icon": "volume",
                        "app": "fs_uae_workspace.apps.hdf_creator_app",
                    },
                    "KeyShow": {
                        "icon": "key-map",
                        "app": "fs_uae_workspace.apps.key_show_app",
                    },
                    "LogViewer": {
                        "icon": "log-viewer",
                        "app": "fs_uae_workspace.apps.logviewer_app",
                    },
                    "TOSECOrganizer": {
                        "icon": "file-manager",
                        "app": "fs_uae_workspace.apps.tosec_organizer_app",
                    },
                    "JoystickConfig": {
                        "icon": "settings",
                        "app": "fs_uae_workspace.apps.joystick_config_app",
                    },
                    "Refresh": {
                        "icon": "settings",
                        "app": "fs_uae_workspace.apps.refresh",
                    },
                    "LockerUploader": {
                        "icon": "settings",
                        "app": "fs_uae_workspace.apps.locker_uploader",
                    },
                },
            },
            "Utilities": {
                "icon": "utilities",
                "children": {
                    "Clock": {
                        "icon": "clock",
                        "app": "fs_uae_workspace.apps.clock_app",
                    },
                },
            },
        },
    },
    "Files": {
        "icon": "volume",
        "children": {
        },
    },
    "Ubuntu": {
        "icon": "ubuntu-volume",
        "children": {
        },
    },
}



DEFAULT_WHDLOAD_VERSION = "18.1"

whdload_support_files = {
    "1ad1b55e7226bd5cd66def8370a69f19244da796":
        "Devs/Kickstarts/kick40068.A1200.RTB",
    "209c109855f94c935439b60950d049527d2f2484":
        "Devs/Kickstarts/kick34005.A500.RTB",
    "973b42dcaf8d6cb111484b3c4d3b719b15f6792d":
        "Devs/Kickstarts/kick40068.A4000.RTB",
    "09e4d8a055b4a9801c6b011e7a3de42bafaf070d":
        "C/uae-configuration",
    "3b40b7277f0408ebb98526205748138f88d84330":
        "C/OSEmu.400",
}

whdload_files = {
    "10.0": {
        "3096b2f41dfebf490aac015bdf0e91a80045c2c0": "C/WHDLoad",
    },
    "13.0": {
        "4bcb393e820d68b0520da9131e0d529018e303d1": "C/WHDLoad",
    },
    "16.0": {
        "883b9e37bc81fc081f78a3f278b732f97bdddf5c": "C/WHDLoad",
    },
    "16.1": {
        "250506c2444d9fb89b711b4fba5d70dd554e6f0e": "C/WHDLoad",
    },
    "16.2": {
        "a8bc2828c7da88f6236a8e82c763c71582f66cfd": "C/WHDLoad",
    },
    "16.3": {
        "5d636899fa9332b7dfccd49df3447238b5b71e49": "C/WHDLoad",
    },
    "16.4": {
        "1bb42fc83ee9237a6cfffdf15a3eb730504c9f65": "C/WHDLoad",
    },
    "16.5": {
        "8974e6c828ac18ff1cc29e56a31da0775ddeb0f0": "C/WHDLoad",
    },
    "16.6": {
        "b268bf7a05630d5b2bbf99616b32f282bac997bf": "C/WHDLoad",
    },
    "16.7": {
        "be94bc3d70d5980fac7fd04df996120e8220c1c0": "C/WHDLoad",
    },
    "16.8": {
        "a3286827c821386ac6e0bb519a7df807550d6a70": "C/WHDLoad",
    },
    "16.9": {
        "b4267a21918d6375e1bbdcaee0bc8b812e366802": "C/WHDLoad",
    },
    "17.0": {
        "0ec213a8c62beb3eb3b3509aaa44f21405929fce": "C/WHDLoad",
    },
    "17.1": {
        "1a907ca4539806b42ad5b6f7aeebacb3720e840d": "C/WHDLoad",
    },
    "2013-03-01": {
        "7ee8516eceb9e799295f1b16909749d08f13d26c": "C/WHDLoad",
    },
    "17.2": {
        "d8f45f7808fb1ac356d88b8848660a6b96f04349": "C/WHDLoad",
    },
    "18.0": {
        "6f778e28673e9f931f81212ab03d9617a41cee40": "C/WHDLoad",
    },
    "18.1": {
        "fb4c64b0b5e682125e53eb2ace9bf0ccd3b8501f": "C/WHDLoad",
    },
}

default_whdload_prefs = """
; wait for button pressed (slave must support this)
;ButtonWait

; disable cache-ability of Chip-Memory
;ChipNoCache

; path for core dump files
;CoreDumpPath=T:

; raw key code to quit with core dump (debug)
;DebugKey=$5b

; command to execute on WHDLoad startup
;ExecuteStartup=rx offline.rexx

; command to execute on WHDLoad exit
;ExecuteCleanup=rx online.rexx

; selects expert mode
;Expert

; raw key code to enter HrtMon/TK
;FreezeKey=$5d

; use MMU (for 68030)
;MMU

; ignore unwanted auto-vector interrupts
;NoAutoVec

; disable audio filter
;NoFilter

; do not flush memory
;NoFlushMem

; do not allocate memory reverse
;NoMemReverse

; disable the disk write cache
;NoWriteCache

;QuitKey=$45

; wait after reading from disk (1/50 seconds)
;ReadDelay=150

; raw key code to restart
;RestartKey=$5c

; command for Show Regs
;ShowRegs=SYS:Utilities/MuchMore W WL=80 WT=80 WW=582 WH=700

; time to display splash window (1/50 seconds)
;SplashDelay=0

; wait after saving something to disk (1/50 seconds)
;WriteDelay=150
"""


from fsui.qt import Qt, QWidget, QLabel, QHBoxLayout


class TitleBarWidget(QWidget):

    def __init__(self, parent, window_handler):
        super().__init__(parent)
        self.setFixedHeight(28)
        self.window_handler = window_handler

        self.setAutoFillBackground(True)
        palette = self.palette()
        palette.setColor(self.backgroundRole(), Qt.white)
        self.setPalette(palette)

        self.caption = Caption(self)
        self.caption.setFixedHeight(self.height())
        # self.caption.setText("FS-UAE Workspace")
        self.caption.setText("Workspace")

        self.caption.setStyleSheet(
            "font-family: \"Open Sans\"; font-size: 10pt; font-weight: bold;")

        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.layout.addSpacing(10)
        self.layout.addWidget(self.caption)

        self.active_window = None
        self.window_handler.window_activated.connect(self.__window_activated)
        self.window_handler.window_changed.connect(self.__window_changed)

    def __window_activated(self, window):
        print("__window_activated", window)
        self.active_window = window
        if self.active_window:
            if self.active_window.is_desktop():
                self.set_caption("Workspace")
            else:
                self.set_caption(self.active_window.name)
        else:
            self.set_caption("")

    def __window_changed(self, window):
        print("__window_changed", window, window.changed)
        if window != self.active_window:
            return
        if window.changed == "name":
            self.set_caption(window.name)

    def set_caption(self, name):
        # self.caption.setText("Applications    Places    System    " + name)
        self.caption.setText(name)


class Caption(QLabel):

    def __init__(self, parent):
        super().__init__(parent)
        self.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        # self.setAlignment(Qt.AlignLeft | Qt.AlignTop)

        # self.setAutoFillBackground(True)
        # palette = self.palette()
        # palette.setColor(self.backgroundRole(), Qt.red)
        # self.setPalette(palette)



import fsui as fsui
from ..I18N import gettext
from .SetupPanel import SetupPanel


class SetupDialog(fsui.Dialog):

    def __init__(self, parent):
        super().__init__(parent, gettext("Import Kickstarts"))
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        buttons.create_close_button()
        self.panel = SetupPanel(self)
        self.panel.set_min_width(620)
        layout.add(self.panel)
